import logging
from qaviton.utils.operating_system import s


format = '%(asctime)s:{%(levelname)s}:%(message)s'


class log_levels:
    CRITICAL = logging.CRITICAL
    FATAL = logging.FATAL
    ERROR = logging.ERROR
    WARNING = logging.WARNING
    WARN = logging.WARN
    INFO = logging.INFO
    DEBUG = logging.DEBUG
    NOTSET = logging.NOTSET


def get_logger(
        log_file=None,
        log_level=log_levels.DEBUG,
        log_name=__name__,
        format=format,
        mode='w',
        log_to_console=True):

    # create logger
    logger = logging.getLogger(log_name)
    logger.setLevel(log_level)
    formatter = logging.Formatter(format)

    if log_to_console is True:
        # create console handler
        handler = logging.StreamHandler()
        # set handler log level
        handler.setLevel(log_level)
        # set handler formatter
        handler.setFormatter(formatter)
        # add handler to logger
        logger.addHandler(handler)

    if log_file is True:
        # create log_file handler
        handler = logging.FileHandler(log_file, mode)
        # set handler log level
        handler.setLevel(log_level)
        # set handler formatter
        handler.setFormatter(formatter)
        # add handler to logger
        logger.addHandler(handler)
    return logger


def get_file_paths_from_log(log):
    file_paths = []
    for handler in log.handlers:
        try:
            file_paths.append(handler.baseFilename)
        except:
            pass
    return file_paths


def get_log_file_directories(log):
    file_paths = get_file_paths_from_log(log)
    log_directory = []
    for file_path in file_paths:
        try:
            log_directory.append(file_path.rsplit(s, 1)[0] + s)
        except:
            pass
    return log_directory



import pytest
import time
from qaviton.utils import unique_id
from qaviton import crosstest
from qaviton.navigator import Navigator
from tests.config.navigation_platforms import platforms
from tests.services.app import App


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigator(platform: crosstest.Platform, request):
    t = time.time()
    test = platform.setup(request)
    print('time to get test platform {}'.format(time.time() - t))

    t = time.time()
    driver = test.driver()
    print('time to get driver {}'.format(time.time() - t))

    t = time.time()
    app = App(driver)
    print('time to get app model {}'.format(time.time() - t))

    # manual navigation connection configuration
    navigate = Navigator(app.google_home)
    navigate.connect_all(
        (app.google_home.navigate_to_GoogleSearchPage, app.google_search),
        (app.google_search.navigate_to_LinkedinHomePage, app.linkedin_home))

    # money time for this feature
    t = time.time()
    navigate.to(app.linkedin_home)
    print('time to get navigation calc {}'.format(time.time() - t))

    t = time.time()
    navigate.perform()
    print('time to get navigation {}'.format(time.time() - t))

    # this is what the tester will want to test
    app.linkedin_home.register.register_and_submit(
        firstname='test123',
        lastname='456',
        email='testmail',
        password='123456gg')

    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigator_with_auto_connect(platform, request):
    app = App.from_platform(platform, request)
    app.navigate.to(app.linkedin_home).perform()
    # this is what the tester will want to test
    app.linkedin_home.register.register_and_submit(
        firstname='test123',
        lastname='456',
        email='testmail',
        password='123456gg')
    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigator_2_navigations(platform, request):
    app = App.from_platform(platform, request)
    app.navigate(app.google_search)
    assert app.navigate.current_page == app.navigate.from_page == app.google_search
    app.navigate(app.linkedin_home)
    assert app.navigate.current_page == app.navigate.from_page == app.linkedin_home
    # this is what the tester will want to test
    app.linkedin_home.register.register_and_submit(
        firstname='test123',
        lastname='456',
        email='testmail',
        password='123456gg')
    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigate_to_self_with_no_possible_navigation(platform, request):
    app = App.from_platform(platform, request)
    app.navigate(app.linkedin_home)
    app.navigate(app.linkedin_home)
    app.linkedin_home.register.reg_submit()
    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigate_with_no_possible_navigation(platform, request):
    app = App.from_platform(platform, request)
    app.navigate(app.linkedin_home)
    # with pytest.raises(PathUnreachableException):
    #     app.navigate(app.google_home)
    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigate_to_self_with_remote_looped_navigation(platform, request):
    app = App.from_platform(platform, request)
    app.navigate(app.google_home)
    app.google_home.google_search_button()
    app.driver.quit()


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_navigate_to_self_with_looped_navigation(platform, request):
    # setup
    app = App.from_platform(platform, request)

    # test flow
    app.navigate(app.google_search)

    # test scenario
    app.navigate(app.google_search)

    # validation
    app.google_search.search_bar.button()

    # teardown
    app.driver.quit()




def find_substring(string, substring, count_start=0, count_stop=0):
    """ find substring with unknown start/finish index
     and return metadata of said string

    :param string: full string (str)
    :param substring: str
    :param count_start: starting index for the search after substring
    :param count_stop: end index for the search after substring
    :return: StringMeta
    """
    # enable reverse counting
    if count_start < 0:
        count_start = len(string) + count_start
    if count_stop < 0:
        count_stop = len(string) + count_stop

    # determine what part of the string to search at
    if count_start < count_stop:
        search_length = count_stop
    else:
        search_length = len(string) - 1

    result = False
    stop_searching_flag = False
    search_start = count_start
    search_stop = search_length

    # start searching from the start point all the way to the search length
    if len(string) >= len(substring):
        for i in range(search_start, search_stop):
            if string[i] == substring[0]:
                if len(substring) == 1:
                    result = True
                    stop_searching_flag = True
                count_start = i
                count_stop = i + 1
                if len(substring) > 1:
                    for b in range(1, len(substring)):
                        if i + b <= search_stop:
                            if string[i + b] != substring[b]:
                                break
                            count_stop += 1
                            if b == len(substring) - 1:
                                result = True
                                stop_searching_flag = True
                                break
                        else:
                            break
                if stop_searching_flag is True:
                    break

    class StringMeta:
        def __init__(self, result, start, stop, next, head, tail, substring, string, search_head, search_tail, search_string):
            """
            :param result: boolean
            :param start: sub_string start index (int)
            :param stop: sub_string last index (int)
            :param next: next index (int)
            :param head: everything before substring (str)
            :param tail: everything after substring (str)
            :param substring: str
            :param string: full string (str)
            :param search_head:
            :param search_tail:
            :param search_string:
            """
            self.result = result
            self.start = start
            self.stop = stop
            self.next = next
            self.head = head
            self.tail = tail
            self.substring = substring
            self.string = string
            self.search_head = search_head
            self.search_tail = search_tail
            self.search_string = search_string

    return StringMeta(
        result=result,
        start=count_start,
        stop=count_stop-1,
        next=count_stop,
        head=string[:count_start],
        tail=string[count_stop:],
        substring=substring,
        string=string,
        search_head=string[search_start:count_start],
        search_tail=string[count_stop:search_stop],
        search_string=string[search_start:search_stop])


def get_string_tail(string, sub_string):
    r = find_substring(string, sub_string)
    if r.result is True:
        return r.tail


def find_string_between_first_two_sub_strings(string, substring1, substring2):
    for i in range(len(string)):
        r = find_substring(string, substring1, i)
        if r.result is True:
            a = r.next
            r = find_substring(string, substring2, a)
            if r.result is True:
                b = r.start
                return string[a:b]
        if i == len(string) - 1:
            return ''


def get_number_from_string(string):
    s = ''
    for c in range(len(string)):
        if string[c].isdecimal():
            s += string[c]
    if len(s) == 0:
        return None
    return int(s)



from qaviton.utils import condition
import pytest


def test_condition_any_in_any1():
    a = "123456789"
    b = "098765"
    assert condition.any_in_any(a, b)


def test_condition_any_in_any2():
    a = ""
    b = "098765"
    assert not condition.any_in_any(a, b)


def test_condition_any_in_any3():
    a = "123456789".split()
    b = "098765"
    assert not condition.any_in_any(a, b)


def test_condition_any_in_any4():
    a = list("123456789")
    b = "098765"
    assert condition.any_in_any(a, b)


def test_condition_any_in_any5():
    a = list("123456789")
    b = list("098765")
    assert condition.any_in_any(a, b)


def test_condition_all_in_any1():
    a = list("123")
    b = list("1234")
    assert condition.all_in_any(a, b)


def test_condition_all_in_any2():
    a = list("12345")
    b = list("1234")
    assert not condition.all_in_any(a, b)


def test_condition_all_in_all1():
    a = list("1234")
    b = list("1234")
    assert condition.all_in_all(a, b)


def test_condition_all_in_all2():
    a = list("1234")
    b = list("12345")
    assert not condition.all_in_all(a, b)


def test_condition_all_in_all3():
    a = list("1234")
    b = list("123")
    assert not condition.all_in_all(a, b)


def test_condition_value_in_many_any():
    class M:
        e1 = 'e1'
        e2 = 'e2'
        e3 = 'e3'

    class V:
        a1 = 'a1'
        a2 = 'a2'
        a3 = 'a3'

    assert condition.value_in_many_any('a1', (vars(M).values(), vars(V).values()))
    assert condition.value_in_many_any('a2', (vars(M).values(), vars(V).values()))
    assert condition.value_in_many_any('a3', (vars(M).values(), vars(V).values()))
    assert condition.value_in_many_any('e1', (vars(M).values(), vars(V).values()))
    assert condition.value_in_many_any('e2', (vars(M).values(), vars(V).values()))
    assert condition.value_in_many_any('e3', (vars(M).values(), vars(V).values()))
    assert not condition.value_in_many_any('e4', (vars(M).values(), vars(V).values()))


def test_condition_any_in_many_any():
    class M:
        e1 = 'e1'
        e2 = 'e2'
        e3 = 'e3'

    class V:
        a1 = 'a1'
        a2 = 'a2'
        a3 = 'a3'

    assert condition.any_in_many_any(('a1','e2'), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('a2', 'e3'), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('a3', 'e1'), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('a1',), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('e2',), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('e2', 'r3'), (vars(M).values(), vars(V).values()))
    assert condition.any_in_many_any(('r3', 'e2'), (vars(M).values(), vars(V).values()))
    assert not condition.any_in_many_any(('e4',), (vars(M).values(), vars(V).values()))
    assert not condition.any_in_many_any(('e4',), (vars(M).values(), vars(V).values()))



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""The WebDriver implementation."""

from appium.webdriver import Remote
from selenium.webdriver.common.action_chains import ActionChains
from qaviton.drivers.common.webdriver import WebDriver as WE
from qaviton.drivers.common.webelement import WebElement


class MobileDriver(WE, Remote):
    """
    Controls a browser by sending commands to a remote server.
    This server is expected to be running the WebDriver wire protocol
    as defined at
    https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol

    :Attributes:
     - session_id - String ID of the browser session started and controlled by this WebDriver.
     - capabilities - Dictionaty of effective capabilities of this browser session as returned
         by the remote server. See https://github.com/SeleniumHQ/selenium/wiki/DesiredCapabilities
     - command_executor - remote_connection.RemoteConnection object used to execute commands.
     - error_handler - errorhandler.ErrorHandler object used to handle errors.
    """

    def __init__(self, command_executor='http://127.0.0.1:4444/wd/hub',
                 desired_capabilities=None, browser_profile=None, proxy=None, keep_alive=False):
        Remote.__init__(self, command_executor, desired_capabilities, browser_profile, proxy, keep_alive)

    def create_web_element(self, element_id):
        """
        Creates a web element with the specified element_id.
        Overrides method and appium Selenium WebDriver in order to always create the qaviton extended WebElement
        """
        return WebElement(self, element_id)



import os


def of(__file__):
    """Returns abs path relative to this file and not cwd
        example:
            from qaviton.utils import path

            path.of(__file__)('../../path/to/file.txt')
    """
    return lambda p: os.path.abspath(os.path.join(os.path.dirname(__file__), p))



from qaviton.utils.string_util import find_string_between_first_two_sub_strings, find_substring, get_number_from_string


def test_string_util():
    string = "dfgfdh FGFHFH fgrkkjhbgvcxc 00050000000701"
    substring = "0005"

    s1 = find_string_between_first_two_sub_strings(string, 'FGH', '007')
    assert s1 == ''
    s1 = find_string_between_first_two_sub_strings(string, 'FGF', '007')
    assert s1 == 'HFH fgrkkjhbgvcxc 000500000' # BUG
    r = find_substring(string, substring)
    n = get_number_from_string(r.substring)
    assert n == 5



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""implementation of common element behavior"""

from selenium.webdriver import Remote
from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.remote.command import Command
from appium.webdriver.webelement import WebElement as WE
from qaviton.drivers.common.functions import WebFunctions
from qaviton.drivers.support import expected_conditions as EC
from qaviton.locator import Locator
from qaviton.heal import Heal


def web_clear(self):
    """override appium and selenium clear method
    in case driver is webbased(selenium).
    Clears the text if it's a text entry element.
    :rtype: WebElement"""
    self._execute(Command.CLEAR_ELEMENT)
    return self


class WebElement(WebFunctions, WE):
    """Represents a DOM element.

        Generally, all interesting operations that interact with a document will be
        performed through this interface.

        All method calls will do a freshness check to ensure that the element
        reference is still valid.  This essentially determines whether or not the
        element is still attached to the DOM.  If this test fails, then an
        ``StaleElementReferenceException`` is thrown, and all future calls to this
        instance will fail."""

    def __init__(self, parent, id_, w3c=False):
        super(self.__class__, self).__init__(parent, id_, w3c)
        if isinstance(parent, Remote):
            self.clear = lambda: web_clear(self)

    def try_to_click(self, timeout=0):
        """ try to click on an element
        :type timeout: int
        :rtype: WebElement | None
        """
        try:
            return self.click(timeout)
        except:
            return None

    def click_at(self):
        ActionChains(self.parent).move_to_element_with_offset(self, self.size["width"]/2, self.size["height"]/2).click().perform()
        return self

    def click(self, locator=None, timeout=0, index=0):
        """ click on an element
        :type locator: tuple(str, str | list[WebElement] | WebElement)
        :type timeout: int
        :type index: int
        :rtype: WebElement
        """
        if locator is None:
            if timeout == 0:
                self._execute(Command.CLICK_ELEMENT)
            else:
                WebDriverWait(self.parent, timeout)\
                    .until(EC.element_to_be_clickable(Locator.element(self), index))._execute(Command.CLICK_ELEMENT)
            return self
        else:
            if timeout == 0:
                element = self.find(locator)
                element.click()
            else:
                element = WebDriverWait(self.parent, timeout).until(EC.element_to_be_clickable(locator, index))
                element._execute(Command.CLICK_ELEMENT)
            return element

    def find_last_children(self, timeout=0):
        """ get the last elements in the tree from the root element
        :type timeout: int
        :rtype: list[WebElement]
        """
        locator = Locator.xpath('./*')
        elements = []
        while True:
            try:
                elements = self.find_all(locator, timeout)
                locator = (locator[0], locator[1] + '/*')
            except:
                return elements

    def find_children(self, timeout=0):
        """ get the last element in the tree from the root element

        :type timeout: int
        :rtype: list[WebElement] | []
        """
        try:
            return self.find_all(Locator.xpath('./*'), timeout)
        except:
            return []

    def drag_to_offset(self, xoffset, yoffset):
        """
        Holds down the left mouse button on the source element,
           then moves to the target offset and releases the mouse button.

        :Args:
         - source: The element to mouse down.
         - xoffset: X offset to move to.
         - yoffset: Y offset to move to.
         :rtype: WebElement
        """
        ActionChains(self.parent).drag_and_drop_by_offset(self, xoffset, yoffset).perform()
        return self

    def hover(self):
        """ hover/move cursor to element
        :rtype: WebElement
        """
        ActionChains(self.parent).move_to_element(self).perform()
        return self

    def hover_and_click(self):
        """ hover/move cursor to element and click
        :rtype: WebElement
        """
        ActionChains(self.parent).move_to_element(self).click(self).perform()
        return self

    def send(self, keys, timeout=0):
        """ click on element, clear element text, send keys to element.
        :type keys: str
        :type timeout: float | int
        :rtype: WebElement
        """
        return self.click(timeout=timeout).clear().send_keys(keys)

    def send_keys(self, *value):
        """Simulates typing into the element.

        :Args:
            - value - A string for typing, or setting form fields.  For setting
              file inputs, this could be a local file path.

        Use this to send simple key events or to fill out form fields::

            form_textfield = driver.find_element_by_name('username')
            form_textfield.send_keys("admin")

        This can also be used to set file inputs.

        ::

            file_input = driver.find_element_by_name('profilePic')
            file_input.send_keys("path/to/profilepic.gif")
            # Generally it's better to wrap the file path in one of the methods
            # in os.path to return the actual path to support cross OS testing.
            # file_input.send_keys(os.path.abspath("path/to/profilepic.gif"))

        """
        WE.send_keys(self, *value)
        return self

    def send_chars(self, chars=''):
        """send individual characters to element.
        this function is useful in flacky applications, where the send keys function can cause issues.
        :type chars: str
        :rtype: WebElement
        """
        for c in chars:
            self.send_keys(c)
        return self

    def get_attributes(self):
        """only for selenium"""
        return self.parent.get_attributes(self)



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
    qaviton implementation of pytest schedualing
    to support parallel testing with
    ordering & dependency features
"""


from xdist.scheduler.load import LoadScheduling
import requests


class QavitonSchedualing(LoadScheduling):
    """Implement load scheduling across nodes.
    This distributes the tests collected across all nodes so each test
    is run just once.  All nodes collect and submit the test suite and
    when all collections are received it is verified they are
    identical collections.  Then the collection gets divided up in
    chunks and chunks get submitted to nodes.  Whenever a node finishes
    an item, it calls ``.mark_test_complete()`` which will trigger the
    scheduler to assign more tests if the number of pending tests for
    the node falls below a low-watermark.
    When created, ``numnodes`` defines how many nodes are expected to
    submit a collection. This is used to know when all nodes have
    finished collection or how large the chunks need to be created.
    Attributes:
    :numnodes: The expected number of nodes taking part.  The actual
       number of nodes will vary during the scheduler's lifetime as
       nodes are added by the DSession as they are brought up and
       removed either because of a dead node or normal shutdown.  This
       number is primarily used to know when the initial collection is
       completed.
    :node2collection: Map of nodes and their test collection.  All
       collections should always be identical.
    :node2pending: Map of nodes and the indices of their pending
       tests.  The indices are an index into ``.pending`` (which is
       identical to their own collection stored in
       ``.node2collection``).
    :collection: The one collection once it is validated to be
       identical between all the nodes.  It is initialised to None
       until ``.schedule()`` is called.
    :pending: List of indices of globally pending tests.  These are
       tests which have not yet been allocated to a chunk for a node
       to process.
    :log: A py.log.Producer instance.
    :config: Config object, used for handling hooks.
    """

    requests.get('https://www.qavitonsaasservice.com/QavitonSchedualing')



from qaviton.locator import Locator
from lxml import etree
from lxml import cssselect
from lxml import html
from io import StringIO


class XMLDriver:
    def __init__(self, xml: str):
        self.xml = xml
        self.parser = None
        self.set_parser(xml)
        try:
            self.tree = etree.parse(StringIO(xml))
        except etree.XMLSyntaxError:
            self.tree = html.fromstring(xml)
            # self.tree = etree.XMLParser(recover=True)

    def set_parser(self, xml: str):
        if "html>" in xml or "<html":
            self.parser = 'HTML'
            etree.HTMLParser()
        else:
            self.parser = 'XML'
            etree.XMLParser()

    def find(self, locator):
        """find elements using xpath
        :param locator: ('XPATH', '//a')
        :type locator: (str, str)
        :rtype: []
        """
        if locator[0] in (Locator.BY.CSS, Locator.BY.CSS_SELECTOR):
            return self.find_by_css(locator[1])
        elif locator[0] != Locator.BY.XPATH:
            locator = Locator.xpath_translator(*locator)
        return self.tree.xpath(locator[1])

    def find_by_xpath(self, xpath: str):
        """find elements using xpath
        :param xpath:
        :rtype: []
        """
        return self.tree.xpath(xpath)

    def find_by_css(self, css: str, namespaces=None, translator='xml'):
        """find elements using css

        To use CSS namespaces, you need to pass a prefix-to-namespace
        mapping as ``namespaces`` keyword argument::

            rdfns = 'http://www.w3.org/1999/02/22-rdf-syntax-ns#'
            select_ns = cssselect.CSSSelector('root > rdf|Description',
                                               namespaces={'rdf': rdfns})

            rdf = etree.XML((
                 '<root xmlns:rdf="%s">'
                   '<rdf:Description>blah</rdf:Description>'
                 '</root>') % rdfns)
            [(el.tag, el.text) for el in select_ns(rdf)]
            [('{http://www.w3.org/1999/02/22-rdf-syntax-ns#}Description', 'blah')]
        :param css:
        :param namespaces:
        :param translator:
        :rtype: []
        """
        return cssselect.CSSSelector(css, namespaces=namespaces, translator=translator)

    def get_text(self, locator):
        """get elements text using xpath
        :param locator: ('XPATH', '//a')
        :rtype: [str]
        """
        return [element.text for element in self.find(locator)]


import pytest
from tests.services.app import App
from tests.config.navigation_platforms import platforms
# from qaviton.fixtures.dependency import Dependencies, depend, dependencies, dependency
# from qaviton.schedualer import QavitonSchedualing
# from qaviton.heal import Heal
# from tests.parameters.locators import locator
# from qaviton.utils import path
#
#
# # the root directory of your project
# workspace = path.of(__file__)('../../')
#
#
# def pytest_configure(config):
#     # adding dependency feature - parallel dependency is exclusive to qaviton cloud
#     # adding self-healing feature - exclusive to qaviton cloud
#     Dependencies.set_path(path.of(__file__)('dependencies'))
#     if not hasattr(config, "slaveinput"):
#         Dependencies.remove_all()
#         Heal.config(workspace, locator)
#         Heal.signal_session_to_start()
#
#
# def pytest_unconfigure(config):
#     if not hasattr(config, "slaveinput"):
#         try:
#             # dependency teardown
#             Dependencies.remove_all()
#             # self healing teardown
#             Heal.signal_session_to_stop()
#         except:
#             pass
#
#
# def pytest_collection_finish(session):
#     # adding dependency & ordering
#     for item in session.items:
#             try:
#                 Dependencies.add(item.name)
#             except:
#                 pass
#
#
# def pytest_xdist_make_scheduler(config, log):
#     # adding parallel ordering feature - exclusive to qaviton cloud
#     return QavitonSchedualing(config, log)
#
#
# @pytest.fixture(scope='session')
# def heal(autouse=True):
#     """this fixture/feature is exclusive to qaviton cloud.
#
#     requirements:
#         your project's file path (workspace)
#         locator class with all your buttons.
#
#     make sure your workspace conveys your project root directory(where .idea - git).
#     the config method takes only one locator class,
#     so make sure to centralize all your locators
#     to 1 locator file with 1 locator class.
#     """
#     Heal.config(workspace, locator)


@pytest.fixture(scope='session', params=platforms.get())
def app(request):
    """fixture for cross-platform model-based application
    always use session scope for this type of fixture & parameterize your platforms
    :rtype: App
    """
    if isinstance(request.param, list) or isinstance(request.param, tuple):
        APPS = []
        for platform in request.param:
            APPS.append(App.from_platform(platform, request))
            request.addfinalizer(APPS[-1].driver.quit)
        return APPS
    APP = App.from_platform(request.param, request)
    request.addfinalizer(APP.driver.quit)
    return APP



import pytest
from qaviton.utils import organize


def test_lists():
    data = []
    r = organize.lists(data)
    assert len(r) == 1

    data = [(123, '5555'), (1, 2, 3, 0, 0), (1, 2), (1, 2, 3, 4)]
    r = organize.lists(data)
    assert len(r) == 80

    data = [(), (1, 2, 3), (1, 2), (1, 2, 3, 4)]
    with pytest.raises(ZeroDivisionError):
        organize.lists(data)


def test_dicts():
    data = {}
    r = organize.dicts(data)
    assert len(r) == 1

    data = dict(a=(123, '5555'), b=(1, 2, 3, 0, 0), c=(1, 2), d=(1, 2, 3, 4))
    r = organize.dicts(data)
    assert len(r) == 80

    data = dict(a=(), b=(1, 2, 3), c=(1, 2), d=(1, 2, 3, 4))
    with pytest.raises(ZeroDivisionError):
        organize.dicts(data)


def test_add():
    item = {"c": (3,)}
    data = [{"a": (1,), "b": (1, 1)}, {"a": (1,), "b": (1, 2)}]

    r = organize.add(data, item)
    assert len(r) == 2

    item = {"c": (3, 4)}
    data = [{"a": (1,), "b": (1, 1)}, {"a": (1,), "b": (1, 2)}]

    r = organize.add(data, item)
    assert len(r) == 4


def test_adds():
    items = [{"c": (3, 4)}, {"d": (5,)}]
    data = [{"a": (1,), "b": (1, 1)}, {"a": (1,), "b": (1, 2)}]

    r = organize.adds(data, items)
    assert len(r) == 4

    items = [{"c": (3, 4)}, {"d": (5, 6)}]
    data = [{"a": (1,), "b": (1, 1)}, {"a": (1,), "b": (1, 2)}]

    r = organize.adds(data, items)
    assert len(r) == 8



from qaviton import crosstest
from tests.config.private import hub
from tests.config.supported_platforms import sessionTimeout

app = 'https://www.google.com/'
screenResolution = "1000x860x24"


# create cross-platform testing object
platforms = crosstest.Platforms()
platforms.web.command_executor = hub


# add chrome browser support
platforms.web({
    'browserName': "chrome",
    'version': "",
    'platform': "ANY",
    'app': app,
    'screenResolution': screenResolution,
    'sessionTimeout': sessionTimeout,
    'enableVNC': True,
    'enableVideo': True,
    'name': "{}",
    'videoName': "{}.mp4",
    'logName': "{}.log"})


# add firefox browser support
platforms.web({
    "browserName": "firefox",
    'version': "61.0",
    "marionette": True,
    "acceptInsecureCerts": True,
    'app': app,
    'screenResolution': screenResolution,
    'sessionTimeout': sessionTimeout,
    'enableVNC': True,
    'enableVideo': True,
    'name': "{}",
    'videoName': "{}.mp4",
    'logName': "{}.log"})



import time
from functools import wraps
from sys import stderr
import traceback


def retry(retries=3, delay=0, backoff=2, exceptions=Exception):
    """Retry function decorator.
    you can also put in delay and backoff to wait between retries
    and define to which exception you want to retry.

    :param retries: number of times to try (not retry) before giving up
    :param delay: initial delay between retries in seconds
    :param backoff: backoff multiplier e.g. value of 2 will double the delay each retry
    :param exceptions: the exception to check. may be a tuple of exceptions to check

    :type retries: int
    :type delay: int
    :type backoff: int
    :type exceptions: Exception or tuple
    """
    def decorator(f):

        @wraps(f)
        def wrapper(*args, **kwargs):
            wait = delay
            for i in range(retries):
                try:
                    return f(*args, **kwargs)
                except exceptions:
                    stderr.write("\nERROR running {}\n{}\nRetries left {}\n".format(f.__name__, traceback.format_exc(), retries-i))
                    time.sleep(wait)
                    wait *= backoff
            return f(*args, **kwargs)
        return wrapper
    return decorator



import pytest
from qaviton.utils.random_util import random_number, random_string
from qaviton.utils import unique_id
from qaviton.utils import condition


@pytest.mark.parametrize('n', [0] * 100, ids=unique_id.id)
def test_random_utils(n):
    n = random_number(1, 100)
    assert 0 < n < 101
    s = "   goGOgoRangers   "
    r = random_string(s, n)
    assert condition.any_in_any(s, r)



import random
from faker import Faker


"""use this to create random names
print(fake.name())
> Rachel Cruz
"""
fake = Faker()


def random_number(minimum, maximum):
    """ get a random number from range
    :type minimum: int
    :type maximum: int
    :rtype: int
    """
    if minimum > maximum:
        maximum = minimum + 1
    numbers = []
    for number in range(minimum, maximum + 1):
        numbers.append(number)
    return random.SystemRandom().choice(numbers)


def no_space_egdes(string, s):
    if s.endswith(' '):
        string = string.replace(" ", "")
        if len(string) == 0:
            raise ValueError("random string cannot end with ' ' space.")

        if s.startswith(' '):
            s = random.SystemRandom().choice(string) + s[1:-1] + random.SystemRandom().choice(string)
        else:
            s = s[:-1] + random.SystemRandom().choice(string)

    elif s.startswith(' '):
        string = string.replace(" ", "")
        if len(string) == 0:
            raise ValueError("random string cannot start with ' ' space.")

        if s.endswith(' '):
            s = random.SystemRandom().choice(string) + s[1:-1] + random.SystemRandom().choice(string)
        else:
            s = random.SystemRandom().choice(string) + s[1:]
    return s


def random_string(string, length=1):
    """ get a random string from a string
    string cannot start/end with ' '
    :type length: int
    :type string: str
    :rtype: str
    """
    return no_space_egdes(string, ''.join(random.SystemRandom().choice(string) for _ in range(length)))


class strings:
    """
    class for different kind of strings
    """
    big_letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
    small_letters = "abcdefghijklmnopqrstuvwxyz"
    numbers = "0123456789"
    space = " "
    invalid_characters = "!@#$%^&*()-_=+`~/?.><\\][}{|"
    special_invalid_characters = "\"'"

    letters = big_letters + small_letters
    letters_and_numbers = letters + numbers
    characters = letters_and_numbers + space
    characters_with_invalid = characters + invalid_characters
    all_characters = characters_with_invalid + special_invalid_characters


class create_random:
    """get a random string
        WARNING: string cannot start/end with ' '(space)
    """
    @staticmethod
    def string(string=strings.characters_with_invalid, length=15):
        """ create random string from a given string.
        if no string is given then random string is created from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789 !@#$%^&*()-_=+`~/?.><\\][}{|

        :param string: string to create from
        :param length: generated random string length
        :rtype: str
        """
        return random_string(string, length)

    @staticmethod
    def all_characters(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789 !@#$%^&*()-_=+`~/?.><\\][}{|"'
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.all_characters, length)

    @staticmethod
    def characters_with_invalid(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789 !@#$%^&*()-_=+`~/?.><\\][}{|
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.characters_with_invalid, length)

    @staticmethod
    def characters(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZ abcdefghijklmnopqrstuvwxyz0123456789
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.characters, length)

    @staticmethod
    def letters_and_numbers(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.letters_and_numbers, length)

    @staticmethod
    def letters(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.letters, length)

    @staticmethod
    def invalid_characters(length=15):
        """ create random string from:
        !@#$%^&*()-_=+`~/?.><\\][}{|
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.invalid_characters, length)

    @staticmethod
    def numbers(length=15):
        """ create random string from:
        0123456789
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.numbers, length)

    @staticmethod
    def small_letters(length=15):
        """ create random string from:
        abcdefghijklmnopqrstuvwxyz
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.small_letters, length)

    @staticmethod
    def big_letters(length=15):
        """ create random string from:
        ABCDEFGHIJKLMNOPQRSTUVWXYZ
        :param length: generated random string length
        :rtype: str
        """
        return random_string(strings.big_letters, length)

    @staticmethod
    def name(length=20):
        """ create random name from fake library:
        :param length: generated random string length
        :rtype: str
        """
        return fake.name()[:length]

    @staticmethod
    def number(minimum=1, maximum=10):
        """ create random number
        :type minimum: int
        :type maximum: int
        :rtype: int
        """
        return random_number(minimum, maximum)

    @staticmethod
    def from_items(items: list, number_of_choices=1):
        """select random items from list
        :param length: generated random string length
        :rtype: []
        """
        items = list(items)
        choices = []
        for i in range(number_of_choices):
            choice = random.SystemRandom().choice(items)
            choices.append(choice)
            items.remove(choice)
        return choices



import pytest
from qaviton.utils.robot import keyboard, mouse


@pytest.mark.skip(reason="this should not run in regression, its a one shot")
def test_robot():
    keyboard.press_and_release('shift+s, space')

    mouse.click()
    mouse.hold()
    mouse.release()

    keyboard.write('The quick brown fox jumps over the lazy dog.')

    # Press PAGE UP then PAGE DOWN to type "foobar".
    keyboard.add_hotkey('page up, page down', lambda: keyboard.write('foobar'))

    # Blocks until you press esc.
    keyboard.wait('esc')

    # Record events until 'esc' is pressed.
    recorded = keyboard.record(until='esc')
    # Then replay back at three times the speed.
    keyboard.play(recorded, speed_factor=3)

    # Type @@ then press space to replace with abbreviation.
    keyboard.add_abbreviation('@@', 'my.long.email@example.com')
    # Block forever.
    keyboard.wait()



import os
import platform


class System:
    name = os.nameplatform_system = platform.system()
    platform_release = platform.release()
    slash = '/'
    if name in ('Windows', 'nt'):
        slash = '\\'


s = System.slash



import pytest
from qaviton.utils.optional_test_action import Optional
from qaviton.utils import condition
from qaviton.utils import unique_id


@pytest.mark.parametrize('n', [0] * 20, ids=unique_id.id)
def test_optional(n):
    class Cat:
        def __init__(self, name):
            self.name = name
            self.optional = Optional()

        def meow(self):
            return self.name + ': meowww..'

        def growl(self):
            return self.name + ': KSSSSssss...'

        def jump(self, msg):
            return self.name + ': has jumped from {}'.format(msg)

        def fight_a_watermelon(self):
            r = [self.name + ': met a watermelon and decided to attack']
            for i in range(self.optional.number(chance=(3, 5))):
                if self.optional(chance=(1, 4)):
                    r.append(self.jump("a sudden blind attack"))
                elif self.optional(chance=(2, 5)):
                    r.append(self.meow())
                if self.optional(chance=(3, 4)):
                    r.append(self.growl())
                elif self.optional():
                    r.append(self.meow())
            return r

    perry = Cat('Perri')
    actions = perry.fight_a_watermelon()

    assert "Perri: met a watermelon and decided to attack" == actions[0]
    if len(actions) > 1:
        assert condition.any_in_any(("Perri: meowww..", "Perri: KSSSSssss...", "Perri: has jumped from a sudden blind attack"), actions)



from qaviton.navigator import Navigator
from tests.pages.components.page import Page
from tests.pages.home import HomePage


class App(Page):
    """use the app page to include all your
    pages/components/api/services/flows
    in a single page application
    """
    def __init__(self, driver, platform):
        Page.__init__(self, driver, platform=platform)
        self.home = HomePage(driver)

        self.navigate = Navigator(self.home, auto_connect=self)



def any_in_any(a, b):
    """return true if any item of 'a' is found
    inside 'b'
    else return false
    """
    if len([x for x in a if x in b]) > 0:
        return True
    else:
        return False


def all_in_any(a, b):
    """return true if every item of 'a' is found
    inside 'b'
    else return false
    """
    if len([x for x in a if x in b]) == len(a):
        return True
    else:
        return False


def all_in_all(a, b):
    """return true if every item of 'a' is matched
    with every item in 'b'
    else return false
    """
    if len([x for x in a if x in b]) == len(a) == len(b):
        return True
    else:
        return False


def value_in_many_any(a, b):
    """return true if item 'a' is found
    inside 'b': a list/tuple of many iterators
    else return false
    """
    for c in b:
        if a in c:
            return True
    return False


def any_in_many_any(a, b):
    """return true if any item of 'a' is found
    inside 'b': a list/tuple of many iterators
    else return false
    """
    for c in b:
        if any_in_any(a, c):
            return True
    return False


def match(a, b):
    """return true if all items of 'a' are matched in 'b', else return false.
    examples:
    match((0,1,2), (0,1,2)) will return True
    match((0,1,2), (0,1,3)) will return False
    """
    if str(a) == str(b):
        return True
    else:
        return False


def relatedclass(o: object, c):
    return issubclass(o.__class__, c)



from qaviton.locator import Locator


class locator(Locator):
    # you can use a variety of methods to build locators
    type_submit = ('type', 'submit')
    qaviton_logo = ('id', 'site-logo')
    qaviton_menu_home_button = ('xpath', '//a[text()="LOGIN"]')
    SIGN_UP_FOR_A_BETA = ('text', 'SIGN UP FOR A BETA')

    # you can use the Locator directly to build your locator
    qaviton_email_demo_request = Locator.xpath('//input[@placeholder="Your Email"]')
    qaviton_company_demo_request = Locator.xpath('//input[@placeholder="Company"]')
    qaviton_name_demo_request = Locator.xpath('//input[@placeholder="Name"]')

    qaviton_send_demo_request = Locator.tuple(
        Locator.xpath('//button[@type="submit"]//*[text()="SEND"]'),
        Locator.index(type_submit, index=0))

    # you can use 'tuple' or 'list' to build a redundant locator and support self healing
    google_search_bar = (
        ('id', 'lst-ib'),
        ('index', 2, ('xpath', '//input')),
        ('xpath', '//input[@type="text"][@value=""]'))

    google_search_button = ('name', 'btnK')
    google_linkedin_search_result = ('text', 'LinkedIn: Log In or Sign Up')

    # id/test-id is the best choice for locators.
    linkedin_reg_firstname = ('id', 'reg-firstname')
    linkedin_reg_lastname = ('id', 'reg-lastname')
    linkedin_reg_email = ('id', 'reg-email')
    linkedin_reg_password = ('id', 'reg-password')
    linkedin_reg_submit = ('id', 'registration-submit')



import sqlite3


class DataBase:
    """
    a very simple implementation
    to create local SQLite database connection
    specified by db_file

    you can integrate your database with your custom commands class
    example usage:
        class command:
            get_users = '''SELECT user_id, password FROM users;'''
            get_user_by_user_id = '''SELECT user_id, password FROM users WHERE user_id = '{}';'''
            get_user_by_id = '''SELECT user_id, password FROM users WHERE id = '{}';'''
            get_last_user = '''SELECT user_id, password FROM users ORDER BY ID DESC LIMIT 1'''
            add_user = '''INSERT INTO users (user_id, password) VALUES ('{}', '{}');'''
            drop_table = '''DROP TABLE IF EXISTS {};'''
            create_users_table = '''CREATE TABLE IF NOT EXISTS users (
                id INTEGER PRIMARY KEY AUTOINCREMENT,
                user_id text,
                password text);'''

        # preferably use 'with' to automatically commit and close the connection
        with DataBase('db') as db:
            db.command = command
            db.execute(db.command.drop_table.format('users'))
            db.execute(db.command.create_users_table)
            db.execute(db.command.add_user.format('id123', 'password1'))

        db = DataBase('db')
        db.command = command
        db.execute(db.command.add_user.format('id777', 'password1'))
        user1 = db.execute(db.command.get_user_by_id.format('id123'))
        user2 = db.execute(db.command.get_user_by_id.format('id777'))
        db.commit()
        db.close()
    """

    def __init__(self, db_file: str, commands=None):
        """ create database connection
        :param db_file: path of db file
        :param commands: custom useful commands
        """
        self.connection = sqlite3.connect(db_file)
        self.cursor = self.connection.cursor()
        self.version = sqlite3.version
        self.command = commands

    def __enter__(self):
        return self

    def __exit__(self, *tp):
        self.commit()
        self.close()

    def __call__(self, command):
        """:rtype: list"""
        return self.execute(command)

    def commit(self):
        self.connection.commit()

    def close(self):
        self.connection.close()

    def execute(self, command):
        """:rtype: list"""
        return self.cursor.execute(command).fetchall()

    def export_table_to_file(self, table, file, titles, permission='w'):
        """ export an entire table to a file

        :param table: table name
        :param file: path to export
        :param titles: table titles
        :param permission: probably write
        """
        self.cursor.execute("select * from {}".format(table))
        table_list = self.cursor.fetchall()
        with open(file, permission) as f:
            f.write(','.join(titles) + '\n')
            for i in table_list:
                s = []
                for a in i:
                    s.append(str(a))
                f.write(','.join(s) + '\n')


import pytest
from qaviton.utils import unique_id

####################################
# you can ignore these tests below #
####################################


@pytest.mark.skip(reason="just testing double parameters for cross-platform & data-driven testing (it's a one shot)")
@pytest.mark.parametrize('test', [0.0,1.0,2.0], ids=unique_id.id)
@pytest.mark.parametrize('data', [0,1,2,3], ids=unique_id.id)
def test_2_layered_data(test, data):
    print(test, data)


@pytest.mark.skip(reason="don't run this its a script")
def test_all_drivers_are_done():
    import psutil

    for proc in psutil.process_iter():
        try:
            if "chrome" in proc.name():
                p = psutil.Process(proc.pid)
                # if 'SYSTEM' not in p.username:
                proc.kill()
        except Exception as e:
            print(e)


from tests.pages.components.page import Page
from tests.config.locators import locator


class YnetHomePage(Page):
    def c3_Hor(self):
        return self.find_all(locator.c3_Hor)

    def evritiframe_1(self):
        return self.find_all(locator.evritiframe_1)

    def multiarticles_15(self):
        return self.find_all(locator.multiarticles_15)

    def multiarticles_5(self):
        return self.find_all(locator.multiarticles_5)

    def close_console(self):
        return self.find_all(locator.close_console)
    
    def su_iframe(self):
        return self.find_all(locator.su_iframe)

    def xButtn(self):
        return self.find_all(locator.xButtn)

    def console_resize(self):
        return self.find_all(locator.console_resize)

    def first_title(self):
        return self.find_all(locator.first_title)

    def arrows(self):
        return self.find_all(locator.arrows)

    def mainSearchSelectText(self):
        return self.find_all(locator.mainSearchSelectText)

    def teaserxnet_1(self):
        return self.find_all(locator.teaserxnet_1)

    def iframe_container(self):
        return self.find_all(locator.iframe_container)

    def null(self):
        return self.find_all(locator.null)

    def ads_300x250_4(self):
        return self.find_all(locator.ads_300x250_4)



from tests.pages.components.page import Page
from tests.config.locators import locator


class LinkedinRegister(Page):
    def reg_firstname(self):
        return self.find(locator.linkedin_reg_firstname)

    def reg_lastname(self):
        return self.find(locator.linkedin_reg_lastname)

    def reg_email(self):
        return self.find(locator.linkedin_reg_email)

    def reg_password(self):
        return self.find(locator.linkedin_reg_password)

    def reg_submit(self):
        return self.find(locator.linkedin_reg_submit)

    def register_and_submit(self, firstname=None, lastname=None, email=None, password=None):
        # first = threaders.thread(self.reg_firstname)
        # last = threaders.thread(self.reg_lastname)
        # mail = threaders.thread(self.reg_email)
        # pss = threaders.thread(self.reg_password)
        #
        # first = first.get_and_join(timeout=10)
        # last = last.get_and_join(timeout=10)
        # mail = mail.get_and_join(timeout=10)
        # pss = pss.get_and_join(timeout=10)
        #
        # first = threading.Thread(target=first.send_keys, args=(firstname,))
        # last = threading.Thread(target=last.send_keys, args=(lastname,))
        # mail = threading.Thread(target=mail.send_keys, args=(email,))
        # pss = threading.Thread(target=pss.send_keys, args=(password,))
        #
        # first.join()
        # last.join()
        # mail.join()
        # pss.join()

        # self.reg_submit().click()
        self.reg_submit()
        # assert threading.active_count() < 3



from qaviton.page import Page as page


class Page(page):
    """component with common behavior to all pages & components alike.
    all of your pages & components should inherit from this common page.

    you should always include here a navigation to your initial login page(starting point)
    so that dependent tests could always start fresh if needed.
    """



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""run qaviton as a service"""

import sys

if len(sys.argv) > 1:
    if sys.argv[1] == 'create':
        from qaviton.scripts import create

        params = []
        tests_dir = 'tests'
        frameworks = ['web']

        if len(sys.argv) > 2:
            frameworks = sys.argv[2].split(',')

            if len(sys.argv) > 3:
                if not sys.argv[3].startswith('--'):
                    tests_dir = sys.argv[3]
                    if len(sys.argv) > 4:
                        params = sys.argv[4:]
                else:
                    params = sys.argv[3:]

        create.framework(frameworks, tests_dir, params)



import uuid
import time
import datetime


def generate():
    """ generate a uuid """
    return str(uuid.uuid4())


def generate_timestamp():
    """ generate a millisecond timestamp uuid """
    t = datetime.datetime.now()
    return int(time.mktime(t.timestamp()) * 1e3 + t.microsecond / 1e3)


"""
generate test id for parameterized tests
    example:
        from qaviton.utils.uuid_generator import uid
        import pytest
        
        @pytest.mark.parametrize('data', [from_zero_to_hero, from_zero_to_hero], ids=uid)  # get test data layer x2
        def test_platforms_and_data(platform: crosstest.Platform, data, request):
            test = platform.setup(request)
"""
id = lambda i: str(i.id)

print(generate())


def lists(data):
    """ organize lists

        example:

        from qaviton.utils import organize

        organize_list = [(1,),(1,2,3),(1,2),(1,2,3,4)]
        print(organize.lists(organize_list))

        will return:

        [[1, 1, 1, 1],[1, 1, 1, 2],[1, 1, 1, 3],[1, 1, 1, 4],[1, 1, 2, 1],[1, 1, 2, 2],
         [1, 1, 2, 3],[1, 1, 2, 4],[1, 1, 3, 1],[1, 1, 3, 2],[1, 1, 3, 3],[1, 1, 3, 4],
         [1, 2, 1, 1],[1, 2, 1, 2],[1, 2, 1, 3],[1, 2, 1, 4],[1, 2, 2, 1],[1, 2, 2, 2],
         [1, 2, 2, 3],[1, 2, 2, 4],[1, 2, 3, 1],[1, 2, 3, 2],[1, 2, 3, 3],[1, 2, 3, 4]]

    :type data: []
    :rtype: []
    """
    try:
        data.sort(key=len)

        all_possibilities = 1
        for i in range(len(data)):
            all_possibilities *= len(data[i])

        organized_data = [[] for _ in range(all_possibilities)]

        looper = all_possibilities
        for data_list in data:
            looper /= len(data_list)
            d = 0
            while d < all_possibilities:
                for i in data_list:
                    for _ in range(int(looper)):
                        organized_data[d].append(i)
                        d += 1
        return organized_data
    except ZeroDivisionError as e:
        raise ZeroDivisionError("empty lists are not allowed") from e


def dicts(data):
    """ organize named lists

        example:

        from qaviton.utils import organize

        organize_dict = {"a":(1,),"b":(1,2,3),"c":(1,2),"d":(1,2,3,4)}
        print(organize.dicts(organize_dict))
        will return:

        [{"a":1, "b":1, "c":1, "d":1},{"a":1, "b":1, "c":1, "d":2},{"a":1, "b":1, "c":1, "d":3},
         {"a":1, "b":1, "c":1, "d":4},{"a":1, "b":1, "c":2, "d":1},{"a":1, "b":1, "c":2, "d":2},
         {"a":1, "b":1, "c":2, "d":3},{"a":1, "b":1, "c":2, "d":4},{"a":1, "b":2, "c":1, "d":1},
         {"a":1, "b":2, "c":1, "d":2},{"a":1, "b":2, "c":1, "d":3},{"a":1, "b":2, "c":1, "d":4},
         {"a":1, "b":2, "c":2, "d":1},{"a":1, "b":2, "c":2, "d":2},{"a":1, "b":2, "c":2, "d":3},
         {"a":1, "b":2, "c":2, "d":4},{"a":1, "b":3, "c":1, "d":1},{"a":1, "b":3, "c":1, "d":2},
         {"a":1, "b":3, "c":1, "d":3},{"a":1, "b":3, "c":1, "d":4},{"a":1, "b":3, "c":2, "d":1},
         {"a":1, "b":3, "c":2, "d":2},{"a":1, "b":4, "c":2, "d":3},{"a":1, "b":3, "c":2, "d":4}]

    :type data: {[]}
    :rtype: [{}]
    """
    try:
        all_possibilities = 1
        for k in data:
            all_possibilities *= len(data[k])

        organized_data = []
        for _ in range(all_possibilities):
            organized_data.append({})
            for k in data:
                organized_data[-1][k] = []

        looper = all_possibilities
        for k in sorted(data, key=lambda k: len(data[k])):
            looper /= len(data[k])
            d = 0
            while d < all_possibilities:
                for i in data[k]:
                    for _ in range(int(looper)):
                        organized_data[d][k].append(i)
                        d += 1
        return organized_data
    except ZeroDivisionError as e:
        raise ZeroDivisionError("empty lists are not allowed") from e


def add(data, item):
    """ add named list to dict
        example:

            from qaviton.utils import organize

            item = {"c":(3,4)}
            data = [{"a":(1,),"b":(1,1)}, {"a":(1,),"b":(1,2)}]

            print(organize.add(data, item))
            will return:

            [{"a":(1,),"b":(1,1),"c":(3,)}, {"a":(1,),"b":(1,1),"c":(4,)},
             {"a":(1,),"b":(1,2),"c":(3,)}, {"a":(1,),"b":(1,2),"c":(4,)}]

        :type item: {[]}
        :type data: {[]}
        :rtype: [{}]
        """
    organized_data = []
    for i in data:
        for d in item:
            for z in item[d]:
                organized_data.append({d:z, **i})
    return organized_data


def adds(data, items):
    for i in items:
        data = add(data, i)
    return data



from selenium.webdriver.common.keys import Keys
from tests.pages.components.page import Page
from tests.config.locators import locator


class GoogleSearchBar(Page):
    def button(self):
        return self.find(locator.google_search_bar)

    def search(self, search):
        self.button().send_keys(search).send_keys(Keys.ENTER)
        self.wait_until_page_loads()



from appium.webdriver.webelement import WebElement
from selenium.webdriver.support.expected_conditions import *
from qaviton.drivers.common.functions import WebFunctions


def _element_if_visible(element, visibility=True):
    """:rtype: WebElement"""
    return element if element.is_displayed() == visibility else False


class presence_of_element_located(object):
    """ An expectation for checking that an element is present on the DOM
    of a page. This does not necessarily mean that the element is visible.
    webLocator - used to find the element
    returns the WebElement once it is located
    :rtype: WebElement
    """
    def __init__(self, locator, index=0):
        self.locator = locator
        self.index = index

    def __call__(self, driver):
        return WebFunctions.find(driver, self.locator, self.index)


class visibility_of_element_located(object):
    """ An expectation for checking that an element is present on the DOM of a
    page and visible. Visibility means that the element is not only displayed
    but also has a height and width that is greater than 0.
    webLocator - used to find the element
    returns the WebElement once it is located and visible
    """
    def __init__(self, locator, index=0):
        self.locator = locator
        self.index = index

    def __call__(self, driver):
        try:
            return _element_if_visible(WebFunctions.find(driver, self.locator, self.index))
        except StaleElementReferenceException:
            return False


class presence_of_all_elements_located(object):
    """ An expectation for checking that there is at least one element present
    on a web page.
    webLocator is used to find the element
    returns the list of WebElements once they are located
    :rtype: list[WebElement]
    """
    def __init__(self, locator):
        self.locator = locator

    def __call__(self, driver):
        return WebFunctions.find_all(driver, self.locator)


class visibility_of_any_elements_located(object):
    """ An expectation for checking that there is at least one element visible
    on a web page.
    webLocator is used to find the element
    returns the list of WebElements once they are located
    """
    def __init__(self, locator):
        self.locator = locator

    def __call__(self, driver):
        """:rtype: list[WebElement]"""
        return [element for element in WebFunctions.find_all(driver, self.locator) if _element_if_visible(element)]


class visibility_of_all_elements_located(object):
    """ An expectation for checking that all elements are present on the DOM of a
    page and visible. Visibility means that the elements are not only displayed
    but also has a height and width that is greater than 0.
    webLocator - used to find the elements
    returns the list of WebElements once they are located and visible
    """
    def __init__(self, locator):
        self.locator = locator

    def __call__(self, driver):
        """:rtype: list[WebElement]"""
        try:
            elements = WebFunctions.find_all(driver, self.locator)
            for element in elements:
                if _element_if_visible(element, visibility=False):
                    return False
            return elements
        except StaleElementReferenceException:
            return False


class text_to_be_present_in_element(object):
    """ An expectation for checking if the given text is present in the
    specified element.
    webLocator, text
    """
    def __init__(self, locator, text_, index=0):
        self.locator = locator
        self.text = text_
        self.index = index

    def __call__(self, driver):
        try:
            element_text = WebFunctions.find(driver, self.locator, self.index).text
            return self.text in element_text
        except StaleElementReferenceException:
            return False


class text_to_be_present_in_element_value(object):
    """
    An expectation for checking if the given text is present in the element's
    webLocator, text
    """
    def __init__(self, locator, text_, index=0):
        self.locator = locator
        self.text = text_
        self.index = index

    def __call__(self, driver):
        try:
            element_text = WebFunctions.find(driver, self.locator, self.index).get_attribute("value")
            if element_text:
                return self.text in element_text
            else:
                return False
        except StaleElementReferenceException:
                return False


class invisibility_of_element_located(object):
    """ An Expectation for checking that an element is either invisible or not
    present on the DOM.

    webLocator used to find the element
    """
    def __init__(self, locator, index=0):
        self.locator = locator
        self.index = index

    def __call__(self, driver):
        try:
            return _element_if_visible(WebFunctions.find(driver, self.locator, self.index), False)
        except (NoSuchElementException, StaleElementReferenceException):
            # In the case of NoSuchElement, returns true because the element is
            # not present in DOM. The try block checks if the element is present
            # but is invisible.
            # In the case of StaleElementReference, returns true because stale
            # element reference implies that element is no longer visible.
            return True


class element_to_be_clickable(object):
    """ An Expectation for checking an element is visible and enabled such that
    you can click it.
    """
    def __init__(self, locator, index=0):
        self.locator = locator
        self.index = index

    def __call__(self, driver):
        """:rtype: WebElement"""
        element = visibility_of_element_located(self.locator, self.index)(driver)
        if element and element.is_enabled():
            return element
        else:
            return False


class element_located_to_be_selected(object):
    """An expectation for the element to be located is selected.
    webLocator is a tuple of (by, path)"""
    def __init__(self, locator, index=0):
        self.locator = locator
        self.index = index

    def __call__(self, driver):
        """:rtype: WebElement"""
        return WebFunctions.find(driver, self.locator, self.index).is_selected()


class element_located_selection_state_to_be(object):
    """ An expectation to locate an element and check if the selection state
    specified is in that state.
    webLocator is a tuple of (by, path)
    is_selected is a boolean
    """
    def __init__(self, locator, is_selected, index=0):
        self.locator = locator
        self.is_selected = is_selected
        self.index = index

    def __call__(self, driver):
        try:
            element = WebFunctions.find(driver, self.locator, self.index)
            return element.is_selected() == self.is_selected
        except StaleElementReferenceException:
            return False



import pytest
from qaviton import crosstest
from qaviton.utils import unique_id
from tests.config.supported_platforms import platforms, sessionTimeout
from qaviton.utils.condition import relatedclass


@pytest.mark.timeout(sessionTimeout*1.5)
@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_models_with_drivers(platform: crosstest.Platform, request):
    test = platform.setup(request)
    driver = test.driver()

    if test.platform["api"] == crosstest.API.WEB:
        driver.get("https://github.com/SeleniumHQ/selenium/wiki/Grid2")

    elif test.platform["api"] == crosstest.API.MOBILE:
        driver.find_element_by_accessibility_id("Add Contact").click()

    driver.quit()


@pytest.mark.timeout(sessionTimeout*1.5)
@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)
def test_new_models_with_drivers(platform, request):
    driver = crosstest.get_driver(platform, request)

    if relatedclass(driver, crosstest.API.WEB):
        driver.get("https://github.com/SeleniumHQ/selenium/wiki/Grid2")

    elif relatedclass(driver, crosstest.API.MOBILE):
        driver.find_element_by_accessibility_id("Add Contact").click()

    driver.quit()



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Exceptions that may happen in the qaviton code.
"""

from selenium.common.exceptions import WebDriverException


class RequiredCapabilitiesException(Exception):
    """
    Thrown when a mandatory capability is missing of configured with invalid value: from DesiredCapabilities.
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class MissingRequiredCapabilitiesException(Exception):
    """
    Thrown when a mandatory capability is missing: from DesiredCapabilities.
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class DriverConnectionException(Exception):
    """
    Thrown when a remote connection to desired driver is not possible due to error with the driver or network failure.
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class DiffException(Exception):
    """
    Thrown when unexpected difference is detected.
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class PathUnreachableException(Exception):
    """
    Thrown when navigation request is impossible.
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class ClickExpectationException(WebDriverException):
    """
    Thrown after click expectation fail.
    an example would be clicking on a button
    while expecting a dom element to be created or deleted but turns out it didn't.
    """

    def __str__(self):
        exception_msg = "element might have been clicked but nothing happened: %s\n" % self.msg
        if self.screen is not None:
            exception_msg += "Screenshot: available via screen\n"
        if self.stacktrace is not None:
            stacktrace = "\n".join(self.stacktrace)
            exception_msg += "Stacktrace:\n%s" % stacktrace
        return exception_msg


class DisabledElementClickException(WebDriverException):
    """
    Thrown after click expectation fail.
    an example would be clicking on a button that should be disabled
    but click action was permitted.
    """

    def __str__(self):
        exception_msg = "element is click-able: %s\n" % self.msg
        if self.screen is not None:
            exception_msg += "Screenshot: available via screen\n"
        if self.stacktrace is not None:
            stacktrace = "\n".join(self.stacktrace)
            exception_msg += "Stacktrace:\n%s" % stacktrace
        return exception_msg


class ElementPresenceException(WebDriverException):
    """
    Thrown when element is unexpectedly present in the dom.
    an example would be navigating to another page
    while expecting a dom element to be deleted but turns out it didn't.
    """

    def __str__(self):
        exception_msg = "element is still present: %s\n" % self.msg
        if self.screen is not None:
            exception_msg += "Screenshot: available via screen\n"
        if self.stacktrace is not None:
            stacktrace = "\n".join(self.stacktrace)
            exception_msg += "Stacktrace:\n%s" % stacktrace
        return exception_msg


class DependencyException(Exception):
    """
    Thrown when test Dependency is missing or yet to have finished running
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


class PageNavigationException(Exception):
    """
    Thrown when page navigation failed
    """

    def __init__(self, msg=None):
        self.msg = msg

    def __str__(self):
        exception_msg = "Message: %s\n" % self.msg
        return exception_msg


def test_linkedin_registration(app):
    app.navigate(app.linkedin_home).register.register_and_submit()


def test_linkedin_registration_again(app):
    app.navigate(app.linkedin_home).register.register_and_submit()


def test_google_search_linkedin(app):
    navigated_page = app.navigate(app.google_search)
    assert app.navigate.current_page == app.navigate.from_page == app.google_search == navigated_page
    navigated_page = app.navigate(app.linkedin_home)
    assert app.navigate.current_page == app.navigate.from_page == app.linkedin_home == navigated_page
    app.linkedin_home.register.register_and_submit()


def test_navigation_works(app):
    app.navigate(app.linkedin_home)
    app.navigate(app.linkedin_home).register.register_and_submit()


def test_qaviton_home(app):
    app.navigate(app.qaviton_home)


def test_google_home(app):
    app.navigate(app.google_home).google_search_button()


def test_qaviton_demo(app):
    app.navigate(app.google_search).search_bar.button()
    app.navigate(app.qaviton_home)
    app.qaviton_home.qaviton_send_demo_request()



def sync_loop(iter1, iter2):
    """let's say you have 2 iterators that go hand to hand, but wait!...
    one of them is a little short on the spot.
    what to do?
    sync-em! run 1 and get the index of the other as well, when the short one is over just re run it from the start.
    works on indexed iterators only
    """
    if len(iter1) > len(iter2):
        big, small = iter1, iter2
    else:
        big, small = iter2, iter1

    i2 = len(small)-1
    for i in range(len(big)):
        if i > i2:
            i2 += len(small)
        if big == iter1:
            yield big[i], small[i-i2+len(small)-1]
        else:
            yield small[i - i2 + len(small) - 1], big[i]



from qaviton import crosstest
from tests.config.secret import hub


web_app = 'https://www.qaviton.com/'
# mobile_app = path.of(__file__)('../../../apps/ContactManager/ContactManager.apk')


screenResolution = "800x600x24"
sessionTimeout = 60


# create cross-platform testing object
platforms = crosstest.Platforms()
platforms.web.command_executor = hub


# add chrome browser support
platforms.web({
    'browserName': "chrome",
    'version': "",
    'platform': "ANY",
    'app': web_app,
    'screenResolution': screenResolution,
    'sessionTimeout': sessionTimeout,
    'enableVNC': True,
    'enableVideo': True,
    'name': "{}",
    'videoName': "{}.mp4",
    'logName': "{}.log"})


# add firefox browser support
platforms.web({
    "browserName": "firefox",
    'version': "61.0",
    "marionette": True,
    "acceptInsecureCerts": True,
    'app': web_app,
    'screenResolution': screenResolution,
    'sessionTimeout': sessionTimeout,
    'enableVNC': True,
    'enableVideo': True,
    'name': "{}",
    'videoName': "{}.mp4",
    'logName': "{}.log"})


# # add internet explorer/opera browser support (not working with selenoid!)
# platforms.web({
#     "browserName": "opera",
#     "version": "",
#     "platform": "ANY",
#     'app': web_app,
#     'screenResolution': screenResolution,
#     'sessionTimeout': sessionTimeout,
#     'enableVNC': True,
#     'enableVideo': True,
#     'name': "{}",
#     'videoName': "{}.mp4",
#     'logName': "{}.log"})


# # add android mobile support
# platforms.mobile({
#     'platformName': "Android",
#     'platformVersion': "6.0",
#     'deviceName': "emulator-5554",
#     'app': mobile_app,
#     'appPackage': 'com.example.android.contactmanager',
#     'appActivity': '.ContactManager',
#     'screenResolution': screenResolution,
#     'sessionTimeout': sessionTimeout,
#     'enableVNC': True,
#     'enableVideo': True,
#     'name': "{}",
#     'videoName': "{}.mp4",
#     'logName': "{}.log"})



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.


from appium.webdriver.webelement import WebElement
from selenium.webdriver.common.by import By
from appium.webdriver.common.mobileby import MobileBy
from qaviton.utils.condition import value_in_many_any


class ByExtension(By):
    """
    This is an extension to the By class
    use this class to choose a locating strategy for web elements!
    """
    ELEMENT = 'element'
    ELEMENTS = 'elements'
    INDEX = 'index'


class ByExtensionMap:
    """
    This extension is intended for mappings in the By class
    """
    CLS = 'class'
    CSS = 'css'
    TEXT = 'text'
    ATTRIBUTE_NAME = 'attribute name'


class Locator:
    """ this is the WebPage element webLocator class
        it is intended to be inherited by other locators of real pages.
        place all common search methods under this super class:

        format method:
            is used only when a paired search (list/tuple that consists of 2 values)
            is required to change the format of its index[1] value.
            example usage:

                import Locator

                class Locator(Locator):
                    button1 = ('id', 'button{}-id')
                    name = ('text', '{name}')

                webLocator = Locator

                print(webLocator.format(webLocator.button1, 1))
                print(webLocator.format(webLocator.common.name, name="louis bow"))

                >> ['id', 'button1-id']
                >> ['text', 'louis bow']

        element method:
            give it an element and get in return a tuple like so:
            (webLocator.element(element))

        text method:
            give it a text and get in return a tuple like so:
            ('text', element)

        __call__ method:
            give it any attribute name & value and get in return a tuple like so:
            (<attribute name string>, <attribute value string>)
    """

    class BY(ByExtension, ByExtensionMap, MobileBy):
        pass

    @staticmethod
    def format(locator, *args, **kw):
        """
        :type locator: tuple
        :rtype: tuple
        """
        if len(locator) > 1:
            if isinstance(locator[1], str):
                return locator[0], locator[1].format(*args, **kw)

    @staticmethod
    def add(locator, value):
        """
        :type locator: tuple
        :type value: str
        :rtype: tuple
        """
        if len(locator) > 1:
            return locator[0], locator[1] + value

    @staticmethod
    def xpath_translator(by, value):
        return By.XPATH, "//*[@{}='{}']".format(by, value)

    @staticmethod
    def any(locator):
        """
        this method maps the element locating strategy to its value
        :type locator: tuple
                example: webLocator=('class', 'top.right')
        """
        by, value = locator

        # match to existing strategies
        if value_in_many_any(by, (vars(ByExtension).values(), vars(By).values(), vars(MobileBy).values())):
            return locator

        # map to extension strategies
        elif by == ByExtensionMap.CLS:
            return Locator.cls(value)
        elif by == ByExtensionMap.CSS:
            return Locator.css(value)
        elif by == ByExtensionMap.TEXT:
            return Locator.text(value)
        elif by == ByExtensionMap.ATTRIBUTE_NAME:
            return Locator.attribute_name(value)

        # find element using any attribute
        else:
            by, value = Locator.xpath_translator(by, value)
        return by, value

    @staticmethod
    def index(locator, index=0):
        """ return an element from a list
        :type locator: tuple
        :param index: what element to return """
        return ByExtension.INDEX, index, locator

    @staticmethod
    def list(*list_of_locators):
        """ return a mapped list of locators
        :type list_of_locators: list or tuple
        :param list_of_locators: a list of locator tuples """
        return [locator for locator in list_of_locators]

    @staticmethod
    def tuple(*tuple_of_locators):
        """ return a mapped tuple of locators
        :type tuple_of_locators: tuple or list
        :param tuple_of_locators: a tuple of locator tuples """
        return tuple_of_locators

    @staticmethod
    def element(web_element):
        """ return a mapped web element
        :type web_element: WebElement """
        return ByExtension.ELEMENT, web_element

    @staticmethod
    def elements(list_of_web_elements):
        """ return a mapped list of web elements
        :type list_of_web_elements: list[WebElement] """
        return ByExtension.ELEMENTS, list_of_web_elements

    @staticmethod
    def attribute_name(attribute_name_locator: str):
        """ return a mapped web element by its attribute name"""
        return By.XPATH, "//*[starts-with(@{},'')]".format(attribute_name_locator)

    @staticmethod
    def text(text_locator: str):
        """ return a mapped web element by its text"""
        return By.XPATH, "//*[{}()='{}']".format(ByExtensionMap.TEXT, text_locator)

    @staticmethod
    def xpath(xpath_locator: str):
        """ return a mapped web element by xpath"""
        return By.XPATH, xpath_locator

    @staticmethod
    def css(css_locator: str):
        """ return a mapped web element by css"""
        return By.CSS_SELECTOR, css_locator

    @staticmethod
    def cls(class_locator: str):
        """ return a mapped web element by its class"""
        return By.CLASS_NAME, class_locator

    @staticmethod
    def id(id_locator: str):
        """ return a mapped web element by its id"""
        return By.ID, id_locator

    @staticmethod
    def link_text(link_text_locator: str):
        """ return a mapped web element by its link text"""
        return By.LINK_TEXT, link_text_locator

    @staticmethod
    def partial_link_text(partial_link_text_locator: str):
        """ return a mapped web element by its partial link text"""
        return By.PARTIAL_LINK_TEXT, partial_link_text_locator

    @staticmethod
    def tag_name(tag_name_locator: str):
        """ return a mapped web element by its html tag name"""
        return By.TAG_NAME, tag_name_locator

    @staticmethod
    def name(name_locator: str):
        """ return a mapped web element by its name property"""
        return By.NAME, name_locator

    @staticmethod
    def ios_predicate(ios_predicate_locator: str):
        """ return a mapped web element by its ios predicate string"""
        return MobileBy.IOS_PREDICATE, ios_predicate_locator

    @staticmethod
    def ios_ui_automation(ios_ui_automation_locator: str):
        """ return a mapped web element by its ios ui automation"""
        return MobileBy.IOS_UIAUTOMATION, ios_ui_automation_locator

    @staticmethod
    def ios_class_chain(ios_class_chain_locator: str):
        """ return a mapped web element by its ios class chain"""
        return MobileBy.IOS_CLASS_CHAIN, ios_class_chain_locator

    @staticmethod
    def android_ui_automation(android_ui_automation_locator: str):
        """ return a mapped web element by its android ui automation"""
        return MobileBy.ANDROID_UIAUTOMATOR, android_ui_automation_locator

    @staticmethod
    def accessibility_id(accessibility_id_locator: str):
        """ return a mapped web element by its accessibility id"""
        return MobileBy.ACCESSIBILITY_ID, accessibility_id_locator

    @staticmethod
    def image(image_locator: str):
        """ return a mapped web element by image"""
        return MobileBy.IMAGE, image_locator






















# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""cross-platform testing implementation"""

from selenium.webdriver.remote.file_detector import UselessFileDetector
from selenium.webdriver.remote.command import Command
from threaders import threaders
from qaviton.drivers.webdriver import WebDriver
from qaviton.drivers.mobile_driver import MobileDriver
from qaviton import settings
from qaviton.exceptions import MissingRequiredCapabilitiesException, DriverConnectionException


# TODO: add more platforms

class API:
    """values need to equal vars(Platforms())"""
    WEB = WebDriver
    MOBILE = MobileDriver
    # IoT = 'iot'
    # CLOUD = 'cloud'
    # PC = 'pc'
    # CODE = 'code'
    # DB = 'db'


class Platforms:
    """all the testing platforms with drivers & desired capabilities"""

    def __init__(self, sessionTimeout: int = 600):
        self.platform_list = []
        self.sessionTimeout = sessionTimeout
        self.web = WebPlatform(self)
        self.mobile = MobilePlatform(self)
        # self.iot = IoTPlatform(self)
        # self.cloud = CloudPlatform(self)
        # self.pc = PCPlatform(self)
        # self.code = CodePlatform(self)
        # self.db = DBPlatform(self)

    def get(self):
        """:rtype: [{}]"""
        return self.platform_list


class _Platform:
    def __init__(self, platforms: Platforms, api):
        self.platforms = platforms
        self.api = api

    def __call__(self, platform):
        """ add platforms to test against
        :param platform: send dictionary of desired capabilities or a list of them
        :type platform: dict or list
        """
        return self.add(platform)

    def add(self, platform):
        if isinstance(platform, list):
            for p in platform:
                self.add(p)
        self.platforms.platform_list.append(Platform(platform))
        return self


class WebPlatform(_Platform):
    """ The WebPlatform is using Selenium Remote Webdriver API:
        Controls a browser by sending commands to a remote server.
        This server is expected to be running the WebDriver wire protocol
        as defined at
        https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol

        :Attributes:
         - session_id - String ID of the browser session started and controlled by this WebDriver.
         - capabilities - Dictionaty of effective capabilities of this browser session as returned
             by the remote server. See https://github.com/SeleniumHQ/selenium/wiki/DesiredCapabilities
         - command_executor - remote_connection.RemoteConnection object used to execute commands.
         - error_handler - errorhandler.ErrorHandler object used to handle errors.
        """

    def __init__(self, platforms: Platforms):
        super(self.__class__, self).__init__(platforms, API.WEB)
        self.command_executor = settings.webdriver_url

    def __call__(self, desired_capabilities, command_executor=None,
                 browser_profile=None, proxy=None,
                 keep_alive=False, file_detector=None, options=None):
        """
        Add a new platform of webdriver configuration that will issue commands using the wire protocol.

            :Args:
                 - app - your app url/file
                 - size - window size; can be 'max', 'min', '800x600' or None for headless/nonesizable browsers
                 - command_executor - Either a string representing URL of the remote server or a custom
                     remote_connection.RemoteConnection object. Defaults to 'http://127.0.0.1:4444/wd/hub'.
                 - desired_capabilities - A dictionary of capabilities to request when
                     starting the browser session. Required parameter.
                 - browser_profile - A selenium.webdriver.firefox.firefox_profile.FirefoxProfile object.
                     Only used if Firefox is requested. Optional.
                 - proxy - A selenium.webdriver.common.proxy.Proxy object. The browser session will
                     be started with given proxy settings, if possible. Optional.
                 - keep_alive - Whether to configure remote_connection.RemoteConnection to use
                     HTTP keep-alive. Defaults to False.
                 - file_detector - Pass custom file detector object during instantiation. If None,
                     then default LocalFileDetector() will be used.
                 - options - instance of a driver options.Options class
        """
        if command_executor is None:
            command_executor = self.command_executor

        platform = locals()
        del platform['self']
        platform['api'] = self.api
        super(self.__class__, self).__call__(platform)


class MobilePlatform(_Platform):

    def __init__(self, platforms):
        super(self.__class__, self).__init__(platforms, API.MOBILE)
        self.command_executor = settings.mobiledriver_url

    def __call__(self, desired_capabilities, command_executor=settings.mobiledriver_url,
                 browser_profile=None, proxy=None, keep_alive=False):
        """
        Add a new platform of webdriver configuration that will issue commands using the wire protocol.

            :Args:
                 - app - your app url/file
                 - size - window size; can be 'max', 'min', '800x600' or None for headless/nonesizable browsers
                 - command_executor - Either a string representing URL of the remote server or a custom
                     remote_connection.RemoteConnection object. Defaults to 'http://127.0.0.1:4444/wd/hub'.
                 - desired_capabilities - A dictionary of capabilities to request when
                     starting the browser session. Required parameter.
                 - browser_profile - A selenium.webdriver.firefox.firefox_profile.FirefoxProfile object.
                     Only used if Firefox is requested. Optional.
                 - proxy - A selenium.webdriver.common.proxy.Proxy object. The browser session will
                     be started with given proxy settings, if possible. Optional.
                 - keep_alive - Whether to configure remote_connection.RemoteConnection to use
                     HTTP keep-alive. Defaults to False.
        """
        if command_executor is None:
            command_executor = self.command_executor

        platform = locals()
        del platform['self']
        platform['api'] = self.api
        super(self.__class__, self).__call__(platform)


# class IoTPlatform(Platform):
#
#     driver_url = WebPlatform.driver_url
#
#     def __init__(self, platforms):
#         super(self.__class__, self).__init__(platforms, API.IoT)
#
#
# class CloudPlatform(Platform):
#
#     driver_url = WebPlatform.driver_url
#
#     def __init__(self, platforms):
#         super(self.__class__, self).__init__(platforms, API.CLOUD)
#
#
# class PCPlatform(Platform):
#
#     driver_url = WebPlatform.driver_url
#
#     def __init__(self, platforms):
#         super(self.__class__, self).__init__(platforms, API.PC)
#
#
# class CodePlatform(Platform):
#
#     driver_url = WebPlatform.driver_url
#
#     def __init__(self, platforms):
#         super(self.__class__, self).__init__(platforms, API.CODE)
#
#
# class DBPlatform(Platform):
#
#     driver_url = WebPlatform.driver_url
#
#     def __init__(self, platforms):
#         super(self.__class__, self).__init__(platforms, API.DB)


class Platform:
    def __init__(self, platform: dict):

        self.platform = dict(platform)

    def setup(self, request):
        """ use this from inside the test to get a matching
        :param request: pytest fixture request
        :rtype: TestCase
        """
        name = request.node.name
        self.platform["desired_capabilities"]["name"] = name
        self.platform["desired_capabilities"]["logName"] = name + '.log'
        self.platform["desired_capabilities"]["videoName"] = name + '.mp4'
        self.platform["desired_capabilities"]["enableVNC"] = True
        self.platform["desired_capabilities"]["enableVideo"] = True
        for k in ('app', 'sessionTimeout', 'screenResolution'):
            if k not in self.platform["desired_capabilities"].keys():
                raise MissingRequiredCapabilitiesException("missing capability: "+k)
        return TestCase(self.platform)


class TestCase:
    def __init__(self, platform):
        self.platform = platform
        self.name = platform["desired_capabilities"]["name"]

    def driver(self, platform=None, connection_timeout=None, page_load_timeout=None, retry=None):
        """return desired driver
        if platform is a list of platforms then you can use the drivers function

        platform = {desired_capabilities, app, size, command_executor,
                 browser_profile, proxy, keep_alive, file_detector, options}

        :param platform: dictionary with:
                    command_executor: it's the remote driver url: default='http://127.0.0.1:4444/wd/hub'
                    desired_capabilities: {"browserName": "chrome", "version": "67.0", "platform": "WINDOWS"}
                    browser_profile=None, proxy=None, keep_alive=False, file_detector=None, options=None
        :param connection_timeout:
        :param page_load_timeout:
        :param retry:
        :rtype: WebDriver or MobileDriver
        """

        if platform is None:
            platform = self.platform

        if platform['api'] == API.WEB:
            # set retry
            if retry is None:
                retry = settings.webdriver_retries

            # set page load timeout
            if connection_timeout is None:
                connection_timeout = settings.webdriver_connection_timeout

            # get remote connection
            for r in range(retry+1):
                try:
                    # from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
                    # driver: WebDriver = platform['api'](
                    #     command_executor=platform['command_executor'],
                    #     desired_capabilities=platform['desired_capabilities'],
                    #     browser_profile=platform['browser_profile'],
                    #     proxy=platform['proxy'],
                    #     keep_alive=platform['keep_alive'],
                    #     file_detector=platform['file_detector'],
                    #     options=platform['options'])
                    # get driver from thread with timeout for connection
                    driver: WebDriver = threaders.thread(
                        platform['api'],
                        command_executor=platform['command_executor'],
                        desired_capabilities=platform['desired_capabilities'],
                        browser_profile=platform['browser_profile'],
                        proxy=platform['proxy'],
                        keep_alive=platform['keep_alive'],
                        file_detector=platform['file_detector'],
                        options=platform['options']
                    ).get_and_join(timeout=connection_timeout)
                    break
                except Exception as e:
                    if r == retry:
                        raise DriverConnectionException("webdriver connection could not be established") from e

            # set local file detector
            driver.file_detector = UselessFileDetector()

            # set page load timeout
            if page_load_timeout is None:
                page_load_timeout = settings.webdriver_page_load_timeout

            # get app url
            for r in range(retry + 1):
                try:
                    # page load from thread with timeout
                    threaders.thread(driver.execute, Command.GET, {'url': platform["desired_capabilities"]["app"]}).get_and_join(timeout=page_load_timeout)
                    # driver.get(platform["desired_capabilities"]["app"])
                    break
                except Exception as e:
                    if r == retry:
                        driver.quit()
                        raise Exception("url could not be loaded") from e

            # try:
            #     resolution = platform["desired_capabilities"]['screenResolution'].split("x")
            #     # driver.execute_script("window.moveTo(arguments[0], arguments[1]);", 0, 0)
            #     try:
            #         driver.set_window_position(0, 0)
            #     except Exception as e:
            #         print(e)
            #     try:
            #         driver.set_window_size(int(resolution[0]), int(resolution[1]))
            #     except:
            #         driver.maximize_window()
            # except Exception as e:
            #     # driver.quit()
            #     # raise RequiredCapabilitiesException("setting screen size failed: please make sure your screenResolution capability is correct - example: 800x600x24 ") from e
            #     print("cannot set the screen size")

        if platform['api'] == API.MOBILE:
            # set retry
            if retry is None:
                retry = settings.mobiledriver_retries
            # set page load timeout
            if connection_timeout is None:
                connection_timeout = settings.mobiledriver_connection_timeout
            for r in range(retry + 1):
                try:
                    driver: MobileDriver = threaders.thread(
                        platform['api'],
                        command_executor=platform['command_executor'],
                        desired_capabilities=platform['desired_capabilities'],
                        browser_profile=platform['browser_profile'],
                        proxy=platform['proxy'],
                        keep_alive=platform['keep_alive']
                    ).get_and_join(timeout=connection_timeout)
                    break
                except Exception as e:
                    if r == retry:
                        try:
                            driver.quit()
                        except:
                            pass
                        raise DriverConnectionException("webdriver connection could not be established") from e
        return driver

    def drivers(self, platforms=None):
        """return the desired drivers
        :param command_executor:  it's the remote driver url: default='http://127.0.0.1:4444/wd/hub'
        :param desired_capabilities: [{"browserName": "chrome", "version": "67.0", "platform": "WINDOWS"}]
        :type desired_capabilities: [{}]
        :param args: ({"browser_profile": None, "proxy": None, "keep_alive": False, "file_detector": None, "options": None}):
        :type platforms: [{}]
        :rtype: [WebDriver or MobileDriver]
        """
        if platforms is None:
            platforms = self.platform
        drivers = []
        for i in platforms:
            drivers.append(self.driver(i))
        return drivers


# generate test id
id = lambda i: str(i.id)


# get driver from platform
def get_driver(platform: Platform, request):
    return platform.setup(request).driver()


# get drivers from platform
def get_drivers(platform: Platform, request):
    return platform.setup(request).drivers()



import pytest
from qaviton.exceptions import DependencyException
from qaviton.utils import filer
from qaviton.utils.operating_system import s
from qaviton.utils.random_util import random_number
import time


class Dependencies:
    max_wait_period: int = 600
    min_wait_period: float = 0.1
    max_read_tries: int = max_wait_period * 10

    path = None

    running = ''
    passed = '0'
    failed = '1'
    skipped = '2'

    @staticmethod
    def wait_iterations():
        return int(Dependencies.max_wait_period/Dependencies.min_wait_period)

    @staticmethod
    def set_path(path):
        Dependencies.path = path
        filer.create_directory(path)

    @staticmethod
    def add(name):
        filer.create_file(Dependencies.path + s + name)

    @staticmethod
    def get(name):
        for i in range(Dependencies.max_read_tries):
            try:
                with open(Dependencies.path + s + name) as f:
                    return f.read()
            except Exception as e:
                if i+1 == Dependencies.max_read_tries:
                    raise e
                time.sleep(random_number(1, 1000)*0.0001)

    @staticmethod
    def get_all():
        return filer.get_dir_files(Dependencies.path)

    @staticmethod
    def set(name, status):
        for i in range(Dependencies.max_read_tries):
            try:
                with open(Dependencies.path + s + name, 'w+') as f:
                    f.write(str(status))
                    return
            except Exception as e:
                if i+1 == Dependencies.max_read_tries:
                    raise e
                time.sleep(random_number(1, 1000)*0.0001)

    @staticmethod
    def rename(name, new_name):
        filer.rename_file(Dependencies.path + s + name, Dependencies.path + s + new_name)

    @staticmethod
    def add_and_set(name, status):
        Dependencies.add(name)
        Dependencies.set(name, status)

    @staticmethod
    def remove(name):
        filer.delete_file(Dependencies.path + s + name)

    @staticmethod
    def remove_all():
        filer.delete_all_files(Dependencies.path)


class Depend:
    dependencies = Dependencies

    def __init__(self):
        self.behave_when_dependency_fails = pytest.skip
        self.behave_when_dependency_is_skipped = pytest.skip

    def skip_when_dependency_fails(self):
        self.behave_when_dependency_fails = pytest.skip
        return self

    def fail_when_dependency_fails(self):
        self.behave_when_dependency_fails = pytest.fail
        return self

    def skip_when_dependency_is_skipped(self):
        self.behave_when_dependency_fails = pytest.skip
        return self

    def fail_when_dependency_is_skipped(self):
        self.behave_when_dependency_fails = pytest.fail
        return self

    def on(self, *dependencies):
        for dependency in dependencies:

            # check if dependency is registered
            for _ in range(Dependencies.wait_iterations()):
                if dependency not in Dependencies.get_all():
                    time.sleep(Dependencies.min_wait_period)
                else:
                    break

            if dependency in Dependencies.get_all():
                status = Dependencies.get(dependency)

                # check if dependency is still running
                for _ in range(Dependencies.wait_iterations()):
                    if status == Dependencies.running:
                        time.sleep(Dependencies.min_wait_period)
                    else:
                        break

                if status == Dependencies.failed:
                    self.behave_when_dependency_fails("Test dependency {} failed".format(dependency))

                elif status == Dependencies.skipped:
                    self.behave_when_dependency_is_skipped("Test dependency {} skipped".format(dependency))

                elif status == Dependencies.running:
                    raise DependencyException(
                        "Test dependency {} has not yet finished running.\n"
                        "please consider using pytest-ordering or increase Dependencies.max_wait_period.\n"
                        "more info at: https://github.com/ftobia/pytest-ordering"
                        .format(dependency))
            else:
                raise DependencyException(
                    "Test dependency {} is either missing or miss-ordered.\n"
                    "In the miss-ordered case please consider reordering your tests with pytest-ordering.\n"
                    "more info at: https://github.com/ftobia/pytest-ordering"
                    .format(dependency))

    def on_pattern(self, *dependency_patterns):
        for pattern in dependency_patterns:
            pattern_found = False
            for dependency in Dependencies.get_all():

                # # check if dependency is registered
                # for _ in range(Dependencies.wait_iterations()):
                #     if pattern not in Dependencies.get_all():
                #         time.sleep(Dependencies.min_wait_period)
                #     else:
                #         break

                if pattern in dependency or pattern == dependency:
                    pattern_found = True
                    status = Dependencies.get(dependency)

                    if status == Dependencies.running:
                        raise DependencyException("Test dependency {} has not yet finished running.\n"
                                                  "please consider using pytest-ordering.\n"
                                                  "more info at: https://github.com/ftobia/pytest-ordering".
                                                  format(dependency))

                    elif status == Dependencies.failed:
                        self.behave_when_dependency_fails("Test dependency {} failed".format(dependency))

                    elif status == Dependencies.skipped:
                        self.behave_when_dependency_is_skipped("Test dependency {} skipped".format(dependency))

            if pattern_found is False:
                raise DependencyException(
                    "Test dependency with pattern {} is either missing or miss-ordered.\n"
                    "In the miss-ordered case please consider reordering your tests with pytest-ordering.\n"
                    "more info at: https://github.com/ftobia/pytest-ordering"
                    .format(pattern))

    def on_parameterized(self, *dependencies):
        for pattern in dependencies:
            pattern_found = False
            for dependency in Dependencies.get_all():
                if pattern == dependency.split("[")[0]:
                    pattern_found = True
                    status = Dependencies.get(dependency)

                    if status == Dependencies.running:
                        raise DependencyException("Test dependency {} has not yet finished running.\n"
                                                  "please consider using pytest-ordering.\n"
                                                  "more info at: https://github.com/ftobia/pytest-ordering".
                                                  format(dependency))

                    elif status == Dependencies.failed:
                        self.behave_when_dependency_fails("Test dependency {} failed".format(dependency))

                    elif status == Dependencies.skipped:
                        self.behave_when_dependency_is_skipped("Test dependency {} skipped".format(dependency))

            if pattern_found is False:
                raise DependencyException(
                    "Test dependency with pattern {}[ is either missing or miss-ordered.\n"
                    "In the miss-ordered case please consider reordering your tests with pytest-ordering.\n"
                    "more info at: https://github.com/ftobia/pytest-ordering"
                    .format(pattern))


class Dependency:
    dependencies = Dependencies

    def __init__(self, name):
        self.name = name
        # Dependencies.add(name)
        self.depend = Depend()

    def set(self, name):
        Dependencies.rename(self.name, name)
        self.name = name
        return self

    def get(self):
        return self.name


@pytest.fixture()
def dependency(request):
    dep = Dependency(request.node.name)
    yield dep
    if request.node.rep_setup.failed:
        Dependencies.set(dep.name, Dependencies.failed)
    elif request.node.rep_setup.skipped:
        Dependencies.set(dep.name, Dependencies.skipped)
    else:
        if request.node.rep_call.failed:
            Dependencies.set(dep.name, Dependencies.failed)
        elif request.node.rep_call.skipped:
            Dependencies.set(dep.name, Dependencies.skipped)
        else:
            Dependencies.set(dep.name, Dependencies.passed)


@pytest.fixture()
def dependencies():
    return Dependencies


@pytest.fixture()
def depend():
    return Depend()



import pytest
from tests.services.app import App
from tests.config.supported_platforms import platforms


@pytest.fixture(scope='session', params=platforms.get())
def app(request):
    """fixture for cross-platform model-based application
    always use session scope for this type of fixture & parameterize your platforms
    :rtype: App
    """
    if isinstance(request.param, list) or isinstance(request.param, tuple):
        APPS = []
        for platform in request.param:
            APPS.append(App.from_platform(platform, request))
            request.addfinalizer(APPS[-1].driver.quit)
        return APPS
    APP = App.from_platform(request.param, request)
    request.addfinalizer(APP.driver.quit)
    return APP


@pytest.fixture(scope='session', params=platforms.get())
def app2(request):
    """you can use multiple apps or even the same app with different driver sessions in your tests
    :rtype: App
    """
    if isinstance(request.param, list) or isinstance(request.param, tuple):
        APPS = []
        for platform in request.param:
            APPS.append(App.from_platform(platform, request))
            request.addfinalizer(APPS[-1].driver.quit)
        return APPS
    APP = App.from_platform(request.param, request)
    request.addfinalizer(APP.driver.quit)
    return APP



from qaviton.page import Page as page


class Page(page):
    """component with common behavior to all pages & components alike.
    all of your pages & components should inherit from this common page.

    you should always include here a navigation to your initial login page(starting point)
    so that dependent tests could always start fresh if needed.
    """
    def navigate_to_GoogleHomePage(self, weight=10, *args, **kwargs):
        self.driver.get("https://www.google.com/")
        self.wait_until_page_loads()

    def navigate_to_QavitonHomePage(self, weight=3, *args, **kwargs):
        self.driver.get("https://www.qaviton.com/")
        self.wait_until_page_loads()


import requests


class Heal:
    @classmethod
    def config(cls, workspace, locator, heal_by=None, heal_directory='/tmp/heal', signal_to_stop='stop'):
        requests.post(
            'https://www.qavitonsaasservice.com/heal-conf',
            json=dict(
                workspace=workspace,
                locator=locator,
                heal_by=heal_by,
                heal_directory=heal_directory,
                signal_to_stop=signal_to_stop))

    @classmethod
    def signal_session_to_start(cls):
        requests.get('https://www.qavitonsaasservice.com/heal-start')

    @classmethod
    def signal_session_to_stop(cls):
        requests.get('https://www.qavitonsaasservice.com/heal-done')

    @classmethod
    def heal_request(cls, locator, healed_locator):
        requests.post(
            'https://www.qavitonsaasservice.com/heal-loc',
            json=dict(locator=locator, healed_locator=healed_locator))



# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""
Uniform Cost Search
"""

import inspect
import heapq
from qaviton.page import Page
from qaviton.exceptions import PathUnreachableException, PageNavigationException


default_weight = 100
page_attributes = dir(Page)


class PriorityQueue:
    """class implementation of priority queue using heapq module"""
    def __init__(self):
        self._queue = []
        self._index = 0

    def insert(self, item, priority):
        heapq.heappush(self._queue, (priority, self._index, item))
        self._index += 1

    def pop(self):
        return heapq.heappop(self._queue)[-1]

    def is_empty(self):
        return len(self._queue) == 0


class Node:
    """class implementation of a node"""
    def __init__(self, node_object):
        """
        :param uuid: unique id - each node in the graph should possess a distinguished id
        self.neighbors represents weight of edges
        """
        self.neighbors = {}
        self.object = node_object

    def connect(self, node, action, weight):
        """add node connection and neighbor registration.
        :param action: the action/function to take in order to advance in the graph
        :type node: Node
        :param weight: cost to the neighbor
        """
        self.neighbors[str(node.object)] = (node, action, weight)


class Graph:
    """class implementation of a graph directed with weight
    The graph not supports multiple edges with the same extremes
    (same output node and same input node)
    """
    def __init__(self):
        """create an empty dict to describe the graph nodes
        with node.key as keys and node objects as values
        """
        self.nodes = {}
        
    def add_nodes(self, nodes):
        """add nodes to graph
        :param nodes: a list of node_objects
        """
        for node in nodes:
            self.add_node(node)
    
    def add_node(self, node):
        """add node to graph
        :param node: a node_object
        """
        if str(node) not in self.nodes:
            self.nodes[str(node)] = Node(node)

    def connect(self, source, destination, action, weight):
        """checks if keys exists in the graph and connect the nodes
        
        :param source: source node uuid or node_object
        :param destination: destination node uuid or node_object
        :param action: the action/function to take in order to advance in the graph
        :param weight: cost to the neighbor
        """
        self.nodes[str(source)].connect(self.nodes[str(destination)], action, weight)

    def get_neighbor_weight(self, source_id: str, neighbor_id: str):
        return self.nodes[source_id].neighbors[neighbor_id][2]

    def get_neighbors_ids(self, node_id: str):
        """return list of node neighbors uuids"""
        return self.nodes[node_id].neighbors.keys()

    def find_path(self, start, goal):
        if str(start) == str(goal):
            return 0, [], []
        elif str(start) not in self.nodes or str(goal) not in self.nodes:
            raise PathUnreachableException('graph is missing start_node {} or goal_node {} does not exist'.format(start, goal))
        else:
            # UCS uses priority queue, priority is the cumulative cost (smaller cost)
            queue = PriorityQueue()

            # insert initial neighbors (cost, actions, nodes) to priority queue
            for neighbor in self.nodes[str(start)].neighbors.values():
                # each item of queue is a tuple (neighbor cumulative_cost, actions to take, nodes to go through)
                if neighbor[0].object != start:
                    queue.insert((neighbor[2], [neighbor[1]], [neighbor[0]]), neighbor[2])

            reached_goal, cumulative_cost_goal = False, -1

            while not queue.is_empty():
                # pop queue item: tuple(cumulative_cost, [real functions to take], [nodes])
                cost, actions, nodes = queue.pop()

                # if goal has been reached
                if str(nodes[-1].object) == str(goal):
                    reached_goal, cumulative_cost_goal = True, cost
                    break

                # insert all neighbors of current node to priority queue
                for neighbor in nodes[-1].neighbors:
                    # check for loops, don't add neighbors that already exist
                    if nodes[-1].neighbors[neighbor][0] not in nodes and nodes[-1].neighbors[neighbor][0].object != start:
                        cumulative_cost = self.get_neighbor_weight(str(nodes[-1].object), neighbor) + cost
                        new_actions, new_nodes = list(actions), list(nodes)
                        new_actions.append(self.nodes[str(nodes[-1].object)].neighbors[neighbor][1])
                        new_nodes.append(self.nodes[neighbor])
                        # insert a new path item into the priority queue.
                        # a tuple (neighbor cumulative_cost, actions to take, nodes to go through)
                        queue.insert((cumulative_cost, new_actions, new_nodes), cumulative_cost)

            if reached_goal:
                return cumulative_cost_goal, actions, nodes
            else:
                raise PathUnreachableException('path is unreachable')


class Navigator:
    """this is a really cool feature meant to do navigation automatically.
        Warnings:
          - please try not to use any arguments with your navigations to keep the process simple,
            if you must try to use lambda functions or make sure all your navigations can receive any kind of argument,
            for extreme cases just use the pull_path method instead of perform.

          - don't use more than 1 instance per page object,
            it will be painfull for the auto-connect function of the navigator.

            examples:
                ###########
                # imports #
                ###########

                from qaviton.page import Page
                from qaviton.locator import Locator
                from qaviton.drivers.webdriver import WebDriver


                #########
                # pages #
                #########

                class BaddaPage(Page):
                    pass


                class BoogiPage(Page):
                    def goPro_navigate_to_BaddaPage(self, weight=5):
                        self.find(Locator.id('badda')).click(timeout=weight * 3)
                        self.wait_until_page_loads()

                    def navigate_to_BaddaPage(self):
                        self.find(Locator.id('badda')).click(timeout=5)

            example 1

                ###################
                # model based app #
                ###################

                class App(Page):
                    def __init__(self, driver):
                        Page.__init__(self, driver)
                        self.boogi_page = BoogiPage(driver)
                        self.badda_page = BaddaPage(driver)
                        self.navigate = Navigator(self.boogi_page, auto_connect=self)

                # create app with driver (if you wanna try this example make sure to send a real webdriver object to your app instead of 0)
                app = App(0)

                # magic!
                print(app.navigate.graph.nodes)
                print(app.navigate.to(app.badda_page).get_path())

                # make sure to send a real webdriver object to your app (app = App(WebDriver()))
                app.navigate.to(app.badda_page).perform()

            example 2

                ###################
                # model based app #
                ###################

                class App(Page):
                    def __init__(self, driver):
                        Page.__init__(self, driver)
                        self.boogi_page = BoogiPage(driver)
                        self.badda_page = BaddaPage(driver)

                navigate = Navigator(self.boogi_page, auto_connect=self)
                navigate.connect_all(
                    (app.boogi_page.goPro_navigate_to_BaddaPage, badda_page, 5),
                    (app.boogi_page.navigate_to_BaddaPage, badda_page))

                app = App(0)
                print(navigate.graph.nodes)
                print(navigate.to(app.badda_page).get_path())

                # make sure to send a real webdriver object to your app (app = App(WebDriver()))
                navigate.to(app.badda_page).perform()
        """

    def __init__(self, landing_page, auto_connect=None):
        """
        :type landing_page: Page
        :type auto_connect: Page | None
        """
        self.current_page = landing_page
        self.from_page = landing_page
        self.nodes = []
        self.actions = []
        self.cost = None
        self.graph = Graph()
        if auto_connect is not None:
            self.auto_connect(auto_connect)

    def __call__(self, page: Page, *args, **kwargs):
        """:rtype: page"""
        return self.to(page).perform(*args, **kwargs)

    def add_node(self, page):
        """ create a new node from page as a node_object
        and add the new node to the Uniform Cost Search Graph

        :type page: Page
        """
        self.graph.add_node(page)

    def add_nodes(self, *pages):
        """ create new nodes from list with pages as a node_objects
        and add the nodes to the Uniform Cost Search Graph

        :type pages: list[Page] | tuple(Page)
        """
        for page in pages:
            self.add_node(page)

    def connect(self, page_bound_navigation, page_to_navigate, weight=default_weight):
        """ connect 2 nodes with pages as node_objects with a bound page navigation method

        :param page_bound_navigation: Page action to take to advance to the neighbor Page
        :param page_to_navigate: the Page neighbor of the Page with the bound method
        :param weight: navigation cost
        :return:
        """
        self.add_nodes(page_bound_navigation.__self__, page_to_navigate)
        self.graph.connect(page_bound_navigation.__self__, page_to_navigate, page_bound_navigation, weight)

    def connect_all(self, *connections):
        """ create multiple nodes connections with lists of (pages as node_objects with a bound page navigation method)

        :param connections: (page_bound_navigation, page_to_navigate, weight=100),
                            (page_bound_navigation, page_to_navigate, weight=100)
        """
        for connection in connections:
            self.connect(*connection)

    def auto_connect(self, app):
        """parse the app for pages/components with
            def {put any thing here}navigate_to_{exact class name of navigated page}(self, optional_but_not_recommended_args, *args, **kwargs):

            example:
                ###########
                # imports #
                ###########

                from qaviton.page import Page
                from qaviton.locator import Locator
                from qaviton.drivers.webdriver import WebDriver


                #########
                # pages #
                #########

                class BaddaPage(Page):
                    pass


                class BoogiPage(Page):
                    def goPro_navigate_to_BaddaPage(self, weight=5):
                        self.find(Locator.id('badda')).click(timeout=weight * 3)
                        self.wait_until_page_loads()

                    def navigate_to_BaddaPage(self):
                        self.find(Locator.id('badda')).click(timeout=5)


                ###################
                # model based app #
                ###################

                class App(Page):
                    def __init__(self, driver):
                        Page.__init__(self, driver)
                        self.boogi_page = BoogiPage(driver)
                        self.badda_page = BaddaPage(driver)
                        self.navigate = Navigator(self.boogi_page, auto_connect=self)


                # create app with driver (if you wanna try this example make sure to send a real webdriver object to your app instead of 0)
                app = App(0)

                # magic!
                print(app.navigate.graph.nodes)
                print(app.navigate.to(app.badda_page).get_path())

                # make sure to send a real webdriver object to your app (app = App(WebDriver()))
                app.navigate.to(app.badda_page).perform()

            in this example the navigation would carry through
            the goPro_navigate_to_BaddaPage which has a timeout of 15 seconds
            instead of the navigate_to_BaddaPage with only 5 seconds to timeout.
            that's because the default weight for navigation functions is 100.
            your best practice would be passing your navigation methods a weight
            equal to a 1/3 of its timeout or equal to the navigation's median run time.

        :type app: Page
        """
        def find_pages_to_navigate():
            for page_connections in connections:
                for i in range(len(page_connections)):
                    for page in pages.values():
                        if page_connections[i][1] == page.__class__.__name__:
                            page_connections[i] = (page_connections[i][0], page, page_connections[i][2])
                            break

        def find_navigations():
            for page in navigation_attributes:
                connections.append([])
                for i in range(len(navigation_attributes[page])):
                    try:
                        at = getattr(pages[page], navigation_attributes[page][i])
                        class_name = navigation_attributes[page][i].split('navigate_to_')[1]
                        if len(class_name) > 0:
                            sig = inspect.signature(at)
                            if 'weight' in sig.parameters:
                                try:
                                    weight = float(str(sig.parameters['weight']).split('=')[1])
                                except:
                                    weight = default_weight
                            else:
                                weight = default_weight
                            connections[-1].append((at, class_name, weight))
                    except:
                        pass

        def find_all_pages(page):
            if str(page) not in pages:
                pages[str(page)] = page
                navigation_attributes[str(page)] = []
                attributes = [m
                    for m in dir(page)
                    if m not in page_attributes
                    and m not in {'timeout', 'driver', 'url'}]

                for i in range(len(attributes)):
                    if 'navigate_to_' in attributes[i]:
                        navigation_attributes[str(page)].append(attributes[i])
                    else:
                        try:
                            at = getattr(page, attributes[i])
                            if issubclass(at.__class__, Page):
                                find_all_pages(getattr(page, attributes[i]))
                        except:
                            pass

        pages = {}
        navigation_attributes = {}
        connections = []

        find_all_pages(app)
        find_navigations()
        find_pages_to_navigate()

        for c in connections:
            self.connect_all(*c)

    def froM(self, page):
        """navigate from specific page

        :type page: Page
        """
        self.set_from_page(page)
        return self

    def to(self, page):
        """find the best path to navigate to a page

        :type page: Page
        """
        self.cost, self.actions, self.nodes = self.graph.find_path(self.from_page, page)
        return self

    def perform(self, *args, **kwargs):
        """perform navigation actions chain
        navigate from any where to anywhere"""
        try:
            for i in range(len(self.actions)):
                self.actions[i](*args, **kwargs)
                self.current_page = self.nodes[i].object
        except Exception as e:
            raise PageNavigationException("navigation interruption at {}".format(self.current_page)) from e
        finally:
            self.actions = []
            self.cost = None
            self.from_page = self.current_page
        return self.current_page

    def set_from_page(self, page):
        """set the page from which to start the navigation chain"""
        self.from_page = self.graph.nodes[str(page)].object

    def update_current_page(self, page):
        """update the navigator with the application's current page"""
        self.current_page = self.graph.nodes[str(page)].object
        self.set_from_page(page)

    def get_path(self):
        return self.actions

    def reset_path(self):
        self.actions = []

    def pull_path(self):
        """use this method if you want to perform the path
        of navigation actions chain outside of the navigator
        to get a list of the navigation methods
        and reset the navigator's path
        """
        navigations = self.actions
        self.actions = []
        return navigations



import random
from qaviton.utils.random_util import random_number


class Optional:
    """ use this object to insert random behavior on none mandatory operations for exploratory testing.
        example:

            class Cat:
                def __init__(self, name):
                    self.name = name
                    self.optional = Optional()

                def meow(self):
                    print(self.name + ': meowww..')

                def growl(self):
                    print(self.name + ': KSSSSssss...')

                def jump(self, msg):
                    print(self.name + ': has jumped from {}'.format(msg))

                def fight_a_watermelon(self):
                    print(self.name + ': met a watermelon and decided to attack')
                    for i in range(self.optional.number(chance=(3, 5))):
                        if self.optional(chance=(1, 4)):
                            self.jump("a sudden blind attack")
                        elif self.optional(chance=(2, 5)):
                            self.meow()
                        if self.optional(chance=(3, 4)):
                            self.growl()
                        elif self.optional():
                            self.meow()

            perry = Cat('Perri') # by the way she was my favorite cat
            perry.fight_a_watermelon()
    """

    def __init__(self, chance=(1, 2), optional_state=True, mandatory=False):
        self.chance = chance
        self.optional_state = optional_state
        self.mandatory = mandatory

    def __call__(self, chance=None, optional_state=None, mandatory=None):
        """
        call this object to run an optional command
        :param chance: tuple of int pair: (1, 2) -> 1:2
        :type chance: (int, int)
        :param optional_state: boolean
        :param mandatory: boolean
        :return: boolean
        """
        if mandatory is None:
            mandatory = self.mandatory
        if mandatory is True:
            return True

        if optional_state is None:
            optional_state = self.optional_state
        if optional_state is False:
            return False

        if chance is None:
            chance = self.chance
        if chance == 0:
            return False

        if optional_state is True:
            if random.random() < chance[0]/chance[1]:
                return True
            else:
                return False
        return False

    def number(self, chance=None, optional_state=None, mandatory=None):
        """ optionally return a random number from chance
        if option is dismissed return value is 0
        if option is mandatory return value is either 1 or highest value
        :param chance: tuple of int pair: (1, 2) -> 1:2
        :type chance: (int, int)
        :param optional_state: boolean
        :param mandatory: boolean
        :rtype: int
        """
        if mandatory is None:
            mandatory = self.mandatory

        if optional_state is None:
            optional_state = self.optional_state
        elif optional_state is False:
            if mandatory is True:
                optional_state = True
            else:
                return 0

        if chance is None:
            chance = self.chance
        if chance[1] == 0:
            if mandatory is True:
                return 1
            else:
                return 0

        if optional_state is True:
            if mandatory is True:
                return chance[1]
            else:
                return random_number(*chance)
        return 0



from tests.pages.components.page import Page
from tests.pages.components.linkedin_register import LinkedinRegister


class LinkedinHomePage(Page):
    def __init__(self, driver, timeout=None):
        Page.__init__(self, driver)
        self.register = LinkedinRegister(driver)



# test runner
import pytest

# things for the tests
from tests.config.supported_platforms import platforms
from tests.config.count_functions import from_zero_to_hero

# things to test
from qaviton.crosstest import WebDriver
from qaviton.crosstest import MobileDriver
from qaviton import crosstest
from qaviton import settings
from qaviton.utils import unique_id


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)  # test per platform
def test_platforms(platform):
    if platform.platform["api"] == crosstest.API.WEB:
        assert platform.platform["desired_capabilities"]["browserName"] in ("firefox", "chrome", "internet explorer")
    elif platform.platform["api"] == crosstest.API.MOBILE:
        assert platform.platform["desired_capabilities"]['platformName'] in ("Android",)
    else:
        raise Exception("bad testing type value: {}".format(platform.platform["api"]))


@pytest.mark.parametrize('platform', platforms.get(), ids=unique_id.id)  # get test platform layer x4
@pytest.mark.parametrize('data', [from_zero_to_hero, from_zero_to_hero], ids=unique_id.id)  # get test data layer x2
def test_platforms_and_data(platform: crosstest.Platform, data, request):
    test = platform.setup(request)

    if test.platform["api"] == WebDriver:
        assert test.platform["command_executor"] == platforms.web.command_executor
        assert test.platform["desired_capabilities"]["browserName"] in ("firefox", "chrome", "internet explorer")

    elif test.platform["api"] == MobileDriver:
        assert test.platform["command_executor"] == settings.mobiledriver_url
        assert test.platform["desired_capabilities"]['platformName'] in ("Android",)

    else:
        raise Exception("test case object not as expected: {}".format(vars(test)))

    assert data == from_zero_to_hero


# health check
@pytest.mark.critical
def test_all_threads_are_done():
    import threading
    assert threading.active_count() == 1




from tests.pages.components.page import Page
from tests.config.locators import locator
from tests.pages.components.google_search_bar import GoogleSearchBar
from selenium.webdriver.common.keys import Keys


class GoogleHomePage(Page):
    def __init__(self, driver):
        Page.__init__(self, driver)
        self.search_bar = GoogleSearchBar(driver)

    def google_search_button(self):
        return self.find(locator.google_search_button)

    def navigate_to_GoogleSearchPage(self, weight=5, *args, **kwargs):
        try:
            self.search_bar.button().send('search').send_keys(Keys.ENTER)
        except:
            self.get_page("https://www.google.com/search?source=hp&ei=M_"
                          "-PW4X2J8epsgGix4rIAQ&q=sd&oq=sd&gs_l=psy-ab.3."
                          ".35i39k1j0j0i131k1l2j0j0i10i1k1j0l4.1997.2074."
                          "0.3892.3.2.0.0.0.0.174.174.0j1.1.0....0...1c.1"
                          ".64.psy-ab..2.1.174.0...0.gG-PzboHJ8c")
        self.wait_until_page_loads()




from jsondiff import diff
from qaviton.exceptions import DiffException


def check_if_elements_are_identical(lst):
    return not lst or lst.count(lst[0]) == len(lst)


class asserts:
    """ assertions to compare values & find diffs """

    @staticmethod
    def elements_are_equal(*args):
        """ compare string value of 2+ variables, throw exception if expectation is not met """

        if check_if_elements_are_identical(args) is False:
            raise DiffException('values {} are not equal as expected'.format(args))

    @staticmethod
    def elements_are_not_equal(*args):
        """ compare string value of 2+ variables, throw exception if expectation is not met """

        if check_if_elements_are_identical(args) is True:
            raise DiffException('values {} are equal not as expected'.format(args))

    @staticmethod
    def value_is_in_a_list(lst, value):
        if value not in lst:
            raise DiffException("value {} doesn't exist in the list".format(value))

    @staticmethod
    def there_is_no_diff(var1, var2):
        diffs = diff(var1, var2)
        if type(diffs) is dict:
            if len(diffs) > 0:
                raise DiffException('found unexpected diffs {}'.format(diffs))

    @staticmethod
    def there_is_a_diff(var1, var2):
        diffs = diff(var1, var2)
        if type(diffs) is dict:
            if len(diffs) == 0:
                raise DiffException('expected diffs {} not found'.format(diffs))



from tests.pages.components.page import Page
from tests.config.locators import locator


class LinkedinRegister(Page):
    def reg_firstname(self):
        return self.find(locator.linkedin_reg_firstname)

    def reg_lastname(self):
        return self.find(locator.linkedin_reg_lastname)

    def reg_email(self):
        return self.find(locator.linkedin_reg_email)

    def reg_password(self):
        return self.find(locator.linkedin_reg_password)

    def reg_submit(self):
        return self.find(locator.linkedin_reg_submit)

    def register_and_submit(self, firstname='test', lastname='tester', email='tester@qaviton.com', password='123test'):
        self.reg_firstname()
        self.reg_lastname()
        self.reg_email()
        self.reg_password()
        self.reg_submit().click()





# Licensed to the Software Freedom Conservancy (SFC) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The SFC licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.

"""The WebDriver implementation."""

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.remote.command import Command
from qaviton.drivers.common.functions import WebFunctions
from qaviton.drivers.common.webelement import WebElement
from qaviton.drivers.support import expected_conditions as EC


class WebDriver(WebFunctions):
    """
    Controls a browser by sending commands to a remote server.
    This server is expected to be running the WebDriver wire protocol
    as defined at
    https://github.com/SeleniumHQ/selenium/wiki/JsonWireProtocol

    :Attributes:
     - session_id - String ID of the browser session started and controlled by this WebDriver.
     - capabilities - Dictionary of effective capabilities of this browser session as returned
         by the remote server. See https://github.com/SeleniumHQ/selenium/wiki/DesiredCapabilities
     - command_executor - remote_connection.RemoteConnection object used to execute commands.
     - error_handler - errorhandler.ErrorHandler object used to handle errors.
    """

    _web_element_cls = WebElement

    def click(self, locator, timeout=0, index=0):
        """ click on an element
        :type locator: tuple(str, str | list[WebElement] | WebElement)
        :type timeout: int
        :type index: int
        :rtype: WebElement
        """
        if timeout == 0:
            element = self.find(locator)
            element.click()
        else:
            element = WebDriverWait(self, timeout).until(EC.element_to_be_clickable(locator, index))
            element._execute(Command.CLICK_ELEMENT)
        return element

    def quit(self):
        """
        Quits the driver and closes every associated window.

        :Usage:
            driver.quit()
        """
        try:
            self.execute(Command.QUIT)
        except:
            pass
        finally:
            self.stop_client()

    @staticmethod
    def find_element_last_children(element, timeout=0):
        """ get the last elements in the tree from the root element
        :type element: WebElement
        :type timeout: int
        :rtype: list[WebElement] | []
        """
        return element.find_last_children(timeout)

    @staticmethod
    def find_elements_last_children(elements, timeout=0):
        """ get the last elements in the tree from the root element
        :type elements: list[WebElement]
        :type timeout: int
        :rtype: list[WebElement] | []
        """
        children_elements = []
        for element in elements:
            children_elements += element.find_last_children(timeout)
        return children_elements

    @staticmethod
    def find_element_children(element, timeout=0):
        """ get the last element in the tree from the root element
        :type element: WebElement
        :type timeout: int
        :rtype: list[WebElement] | []
        """
        return element.find_children(timeout)

    @staticmethod
    def get_elements_children(elements, timeout=0):
        """ get elements children
        :type timeout: int
        :rtype: list[WebElement] | []
        """
        children_elements = []
        for element in elements:
            children_elements += element.find_children(timeout)
        return children_elements

    @staticmethod
    def drag_to_offset(origin_el, xoffset, yoffset):
        """
        Holds down the left mouse button on the source element,
           then moves to the target offset and releases the mouse button.

        :Args:
         - source: The element to mouse down.
         - xoffset: X offset to move to.
         - yoffset: Y offset to move to.
         :type origin_el: WebElement
         :rtype: WebDriver
        """
        return origin_el.drag_to_offset(xoffset, yoffset)

    @staticmethod
    def hover(element):
        """ hover/move cursor to element
        :type element: WebElement
        :rtype: WebDriver
        """
        return element.hover()

    @staticmethod
    def hover_and_click(element):
        """ hover/move cursor to element and click
        :type element: WebElement
        :rtype: WebDriver
        """
        return element.hover_and_click()
















